# Ruby Lexer implemented with Rockit's parser combinator library
#
# Note that this lexer do not handle multi-byte characters since I haven't
# understood them or how to parse them... ;-) Maybe some day.
#
# Copyright (c) 2001 Robert Feldt, feldt@ce.chalmers.se
# All rights reserved.
#
# This is free software, blah blah blah...
#
require 'rockit/parse/combinator_parsers'
require 'rockit/parse/result_mapping'
require 'rockit/parse/lexer'
require 'rockit/parse/std_library'
include Rockit::Parse

require 'rulator/terms'

module Ruby
  # For now we use an Array to represent tokens since Racc requires an
  # Array of length 2. We cheat a bit since we want it to contain all
  # info sent from lexer to parser.
  class RaccToken < Array
    attr_reader :info, :position

    def initialize(*args)
      super([2,args.length].min)
      args.each_with_index {|v, i| self[i] = v if i < 2}
      @position = args[2] if args.length >= 3
      @info = args[3] if args.length >= 4
    end

    def length
      2 # Racc needs this!
    end

    def token_type
      self[0]
    end

    def lexeme
      self[1]
    end

    def inspect
      if self[0]
	"#{self[0]}(#{self[1].inspect}, #{@position}" +
	  (@info ? ", #{@info.inspect}" : "") + ")"
      else
	"EOF"
      end
    end
  end

  # The Ruby lexer uses states to choose between different tokens. Below we
  # call them modes so that we don't confuse them with the parser state.
  EXPR_BEG    = 0      # ignore newline, +/- is a sign.
  EXPR_END    = 1      # newline significant, +/- is a operator.
  EXPR_ARG    = 2      # newline significant, +/- is a operator.
  EXPR_CMDARG = 3      # newline significant, +/- is a operator. */
  EXPR_ENDARG = 4      # newline significant, +/- is a operator. */
  EXPR_MID    = 5      # newline significant, +/- is a operator.
  EXPR_FNAME  = 6      # ignore newline, no reserved words.
  EXPR_DOT    = 7      # right after `.' or `::', no reserved words.
  EXPR_CLASS  = 8      # immediate after `class', no here document.
  
  # And some often used sets of states:
  EXPR_STATES = 
    [EXPR_BEG, EXPR_END, EXPR_ARG, EXPR_CMDARG, EXPR_ENDARG,
     EXPR_MID, EXPR_FNAME, EXPR_DOT, EXPR_CLASS]
  BEG_AND_MID = [EXPR_BEG, EXPR_MID]
  NOT_BEG_OR_MID = EXPR_STATES - [EXPR_BEG, EXPR_MID]
  FNAME_AND_DOT = [EXPR_FNAME, EXPR_DOT]
  NOT_FNAME_OR_DOT = EXPR_STATES - [EXPR_FNAME, EXPR_DOT]
  END_AND_ARG = [EXPR_END, EXPR_ARG]
  NOT_END_OR_ARG = EXPR_STATES - END_AND_ARG
  END_AND_ENDARG = [EXPR_END, EXPR_ENDARG]
  NOT_BEG_MID_OR_ARG = EXPR_STATES - [EXPR_BEG, EXPR_MID, EXPR_ARG]

  # Reserved words from lex.c in Ruby 1.6.5
  RESERVED_WORDS = {
    "end"      => [:kEND, :kEND, EXPR_END],
    "else"     => [:kELSE, :kELSE, EXPR_BEG],
    "case"     => [:kCASE, :kCASE, EXPR_BEG],
    "ensure"   => [:kENSURE, :kENSURE, EXPR_BEG],
    "module"   => [:kMODULE, :kMODULE, EXPR_BEG],
    "elsif"    => [:kELSIF, :kELSIF, EXPR_BEG],
    "def"      => [:kDEF, :kDEF, EXPR_FNAME],
    "rescue"   => [:kRESCUE, :kRESCUE_MOD, EXPR_MID],
    "not"      => [:kNOT, :kNOT, EXPR_BEG],
    "then"     => [:kTHEN, :kTHEN, EXPR_BEG],
    "yield"    => [:kYIELD, :kYIELD, EXPR_ARG],
    "for"      => [:kFOR, :kFOR, EXPR_BEG],
    "self"     => [:kSELF, :kSELF, EXPR_END],
    "false"    => [:kFALSE, :kFALSE, EXPR_END],
    "retry"    => [:kRETRY, :kRETRY, EXPR_END],
    "return"   => [:kRETURN, :kRETURN, EXPR_MID],
    "true"     => [:kTRUE, :kTRUE, EXPR_END],
    "if"       => [:kIF, :kIF_MOD, EXPR_BEG],
    "defined?" => [:kDEFINED, :kDEFINED, EXPR_ARG],
    "super"    => [:kSUPER, :kSUPER, EXPR_ARG],
    "undef"    => [:kUNDEF, :kUNDEF, EXPR_FNAME],
    "break"    => [:kBREAK, :kBREAK, EXPR_END],
    "in"       => [:kIN, :kIN, EXPR_BEG],
    "do"       => [:kDO, :kDO, EXPR_BEG],
    "nil"      => [:kNIL, :kNIL, EXPR_END],
    "until"    => [:kUNTIL, :kUNTIL_MOD, EXPR_BEG],
    "unless"   => [:kUNLESS, :kUNLESS_MOD, EXPR_BEG],
    "or"       => [:kOR, :kOR, EXPR_BEG],
    "next"     => [:kNEXT, :kNEXT, EXPR_END],
    "when"     => [:kWHEN, :kWHEN, EXPR_BEG],
    "redo"     => [:kREDO, :kREDO, EXPR_END],
    "and"      => [:kAND, :kAND, EXPR_BEG],
    "begin"    => [:kBEGIN, :kBEGIN, EXPR_BEG],
    "__LINE__" => [:k__LINE__, :k__LINE__, EXPR_END],
    "class"    => [:kCLASS, :kCLASS, EXPR_CLASS],
    "__FILE__" => [:k__FILE__, :k__FILE__, EXPR_END],
    "END"      => [:klEND, :klEND, EXPR_END],
    "BEGIN"    => [:klBEGIN, :klBEGIN, EXPR_END],
    "while"    => [:kWHILE, :kWHILE_MOD, EXPR_BEG],
    "alias"    => [:kALIAS, :kALIAS, EXPR_FNAME],
  }

  # We create a special ParseState with a global mode attribute.
  class ModedState < OptimizedParseState
    attr_accessor :mode

    def initialize(*args)
      super(*args)
      @mode = nil
    end
  end

  # We introduce special parsers for checking and setting the mode
  class SetModeParser < Parser
    def initialize(mode)
      @mode = mode
      super()
    end

    def match(aState)
      aState.mode = @mode
      true
    end
  end

  class CheckModeParser < Parser
    def initialize(modes)
      super()
      @modes = init_modes(modes)
    end

    def init_modes(modes)
      if Hash === modes
	modes
      elsif Array === modes
	h = Hash.new
	modes.each {|m| h[m] = true}
	h
      else
	{modes => true}
      end
    end

    def match(aState)
      @modes[aState.mode]
    end
  end

  # Convenience methods for creating mode setting/getting parsers
  def Ruby.set_modep(aParser, toMode)
    seq(aParser, SetModeParser.new(toMode))
  end

  def Ruby.check_modep(aParser, fromModes)
    seq(CheckModeParser.new(fromModes), aParser)
  end

  def Ruby.modep(aParser, fromModes = [EXPR_BEG], toMode = nil)
    sub_parsers = [CheckModeParser.new(fromModes), aParser]
    sub_parsers.push SetModeParser.new(toMode) if toMode
    seq(*sub_parsers)
  end

  # A RaccToken creating result mapper
  class MakeRaccToken < ResultMappingFunction
    def initialize(tokenType, info = nil)
      @token_type, @info = tokenType, info
    end

    def call(aParseResult, aParseState = nil, aPosition = nil)
      RaccToken.new(@token_type, aParseResult, 
		    position(aParseResult, aParseState, aPosition), 
		    @info)
    end

    def position(aParseResult, aState = nil, aPosition = nil)
      if aPosition
	aPosition
      elsif aState && String === aParseResult
	aState.position - aParseResult.length
      else
	nil
      end
    end
  end

  class MakeToken < MakeRaccToken
    def initialize(tokenClass, info = nil)
      @token_class, @info = tokenClass, info
    end

    def call(aParseResult, aParseState = nil, aPosition = nil)
      tok = @token_class[aParseResult, @info]
      tok.position = position(aParseResult, aParseState, aPosition)
      tok
    end
  end

  module TokenGrammar
    # We move methods in here for easier reference
    # Convenience methods for creating mode setting/getting parsers
    def TokenGrammar.set_modep(*args); Ruby.set_modep(*args); end
    def TokenGrammar.check_modep(*args); Ruby.check_modep(*args); end
    def TokenGrammar.modep(*args); Ruby.modep(*args); end

    # And here's the method for creating a parser that returns a Racc token
    def TokenGrammar.rbtok(o, tokenType, info = nil, setState = nil)
      p = make_parser(o)
      p = set_modep(p, setState) if setState
      p.apply(MakeRaccToken.new(tokenType, info))
    end

    def TokenGrammar.rbtok2(o, tokenType, info = nil, setState = nil)
      p = make_parser(o)
      p = set_modep(p, setState) if setState
      p.apply(MakeToken.new(tokenType, info))
    end

    # Short hand for creating a 
    #   set_modep( choice(rbtok(...), ..., rbtok(...)), EXPR_BEG ) parser
    #
    def TokenGrammar.mrbtp(*tokSpecs)
      rbtoks = []
      tokSpecs.each do |string, tokenType, info|
	rbtoks.push rbtok(string, tokenType, info)
      end
      set_modep(choice(*rbtoks), EXPR_BEG)
    end

    #========================================================================
    # Start of TokenGrammar - the specification of a parser for Ruby tokens
    # 
    # Note that some of the definitions below are order-sensitive since they
    # use the choice parser combinator which tries parsers in order.
    #

    Equality = mrbtp( ["===", :tEQQ],   ["==", :tEQ], ["=~",  :tMATCH], 
		      ["=>",  :tASSOC], ["=",   "="] )

    Star = choice(rbtok("**=", :tOP_ASGN, :tPOW),
		  rbtok("**",  :tPOW),
		  rbtok("*=",  :tOP_ASGN, "*"),
		  modep(rbtok("*",   :tSTAR), BEG_AND_MID,     EXPR_BEG),
		  modep(rbtok("*",   "*"),    NOT_BEG_OR_MID, EXPR_BEG))

    Exclamation = mrbtp( ["!=", :tNEQ], ["!~", :tNMATCH], ["!", "!"] )

    class EscapeSequenceToInt < ResultMappingFunction
      @@map = {
	"\\n" => ?\n, "\\t" => ?\t, "\\r" => ?\r, "\\f" => ?\f,
	"\\v" => ?\v, "\\a" => ?\a, "\\e" => ?\e, "\\b" => ?\b,
	"\\s" => ?\s
      }
	
      def call(aParseResult, aParseState = nil, aPosition = nil)
	if (res = @@map[aParseResult])
	  res
	else
	  if ?x == aParseResult[1]
	    aParseResult[2..-1].hex
	  else
	    aParseResult[1..-1].oct
	  end
	end
      end
    end    
    
    # This is not complete since I haven't understood all of the more
    # unusual escape sequences.
    EscapeSequence = 
      seq("\\", 
	  choice(one_of("ntrfvaebs"),
		 seq("x", repeat(HexDigit, 1, 2)),
		 repeat(Digit, 1, 3))
	  ).apply(EscapeSequenceToInt.new)

    CharValue = 
      choice(check_modep(rbtok(look_ahead(" "), "?"), EXPR_ARG),
	     rbtok2(EscapeSequence, IntegerLit),
	     rbtok2(AnyChar.apply(CharToIntFunction.new), IntegerLit)
	     ).or_raises("incomplete character syntax")

    QuestionMark = 
      choice(modep(rbtok("?", "?"), [EXPR_END, EXPR_ENDARG], EXPR_BEG),
	     modep(seq("?", CharValue).lift(1), EXPR_STATES - END_AND_ENDARG,
		   EXPR_END))

    IdentChar = symbols("a".."z", "A".."Z", "_", "0".."9")

    # White space that is not a newline
    WhiteSpace = symbols(" ", "\t", "\f", "\v", "\r", "\13") #\13

    # Create a parser that is preceded by WhiteSpace
    def TokenGrammar.wsp(aParser)
      seq(look_back(WhiteSpace), aParser)
    end

    HereDocTerminator = capture(plus(IdentChar), :heredoc_terminator)

    def TokenGrammar.quoted_heredoc_header(quote)
      seq( quote, HereDocTerminator, capture(quote, :heredoc_type), 
	   RestOfLine )
    end

    HereDocHeader = 
      choice( *[succeed, "\"", "'", "`"].map {|q| quoted_heredoc_header(q)} )

    def TokenGrammar.until_eohd(indent)
      l = [(mult(one_of(" \t")) if indent), ref(:heredoc_terminator)]
      lines_until( seq( skip( seq( *l.compact ) ), look_ahead(maybe(Newline))))
    end

    HereDoc = 
      seq( skip("<<"),
	   choice( seq( skip("-"), skip(HereDocHeader), until_eohd(true) ),
		   seq( skip(HereDocHeader), until_eohd(false) ) 
		  )
	  )

    class HereDocToString < MakeRaccToken
      def initialize(key)
	@attribute_key = key
      end

      def call(aParseResult, aParseState = nil, aPosition = nil)
	case aParseState.get_attribute(@attribute_key)
	when "", "\""
	  RaccToken.new(:tDSTRING, aParseResult, 
			position(aParseResult, aParseState, aPosition))
	when "'", "`"
	  RaccToken.new(:tSTRING, aParseResult, 
			position(aParseResult, aParseState, aPosition))
	end
      end
    end

    HereDocument = 
      choice( modep( HereDoc, 
		     EXPR_STATES - [EXPR_END, EXPR_CLASS, EXPR_ARG], EXPR_END),
	      modep( wsp(HereDoc), EXPR_ARG, EXPR_END )
	     ).apply(HereDocToString.new(:heredoc_type))
    
    LessThan =
      choice( HereDocument,
	      set_modep(choice(rbtok("<=>", :tCMP),
			       rbtok("<=",  :tLEQ),
			       rbtok("<<=", :tOP_ASGN, :tLSHFT),
			       rbtok("<<",  :tLSHFT),
			       rbtok("<",   "<") ),
			EXPR_BEG)
	     )

    GreaterThan =
      mrbtp( [">=", :tGEQ], [">>=", :tOP_ASGN, :tRSHFT], 
	     [">>", :tRSHFT], [">", ">"] ) 

    Ampersand =
      set_modep( choice( rbtok("&&=", :tOP_ASGN, :tANDOP),
			 rbtok("&&",  :tANDOP),
			 rbtok("&=",  :tOP_ASGN, "&"),
			 check_modep(rbtok(wsp("&"), :tAMPER), EXPR_ARG),
			 check_modep(rbtok("&",      "&"),     EXPR_ARG),
			 check_modep(rbtok("&",      :tAMPER), BEG_AND_MID),
			 check_modep(rbtok("&",      "&"), NOT_BEG_MID_OR_ARG)
			),
		 EXPR_BEG
		)

    Bar = mrbtp( ["||=", :tOP_ASGN, :tOROP], ["||", :tOROP],
		 ["|=",  :tOP_ASGN, "|"],    ["|", "|"] )  

    Plus =
      choice( check_modep(rbtok("+@", :tUPLUS), FNAME_AND_DOT),
	      check_modep(rbtok("+",  "+"),      FNAME_AND_DOT),
	      modep(rbtok("+=", :tOP_ASGN, "+"), NOT_FNAME_OR_DOT, EXPR_BEG),
	      modep(rbtok(string("+").not_followed_by(Digit), :tUPLUS),
		    BEG_AND_MID, EXPR_BEG),
	      modep(rbtok(seq(wsp("+"), look_ahead(WhiteSpace)), "+"),
		    EXPR_ARG, EXPR_BEG),
	      
	     # Error with this one:
             # modep(rbtok(wsp("+").not_followed_by(Digit), :tUMINUS),   
	     # EXPR_ARG, EXPR_BEG),

	     # this one should match also if in EXPR_ARG but not preceded by
	     # whitespace:
	      modep(rbtok("+", "+"), 
		    EXPR_STATES - BEG_AND_MID - FNAME_AND_DOT - [EXPR_ARG], 
		    EXPR_BEG) )


    Minus =
      choice( check_modep(rbtok("-@", :tUMINUS), FNAME_AND_DOT),
	      check_modep(rbtok("-",  "-"),      FNAME_AND_DOT),
	      modep(rbtok("-=", :tOP_ASGN, "-"), NOT_FNAME_OR_DOT, EXPR_BEG),
	      modep(rbtok(string("-").not_followed_by(Digit), :tUMINUS),
		    BEG_AND_MID, EXPR_BEG),
	     # Error with this one:
             # modep(rbtok(wsp("-").not_followed_by(Digit), :tUMINUS),   
	     # EXPR_ARG, EXPR_BEG),

	      modep(rbtok(seq(wsp("-"), look_ahead(WhiteSpace)), "-"),
		    EXPR_ARG, EXPR_BEG),

	     # this one should match also if in EXPR_ARG but not preceded by
	     # whitespace:
	      modep(rbtok("-", "-"), 
		    EXPR_STATES - BEG_AND_MID - FNAME_AND_DOT - [EXPR_ARG], 
		    EXPR_BEG) )

    Caret = mrbtp( ["^=", :tOP_ASGN, "^"], ["^", "^"] )

    Tilde = choice(modep(rbtok(seq("~", skip("@")), "~"), 
			 FNAME_AND_DOT, EXPR_BEG),
		   rbtok("~", "~", nil, EXPR_BEG))

    Slash = 
      choice(modep(rbtok(seq("/", skip("=")),  :tOP_ASGN, "/"),   
		   NOT_BEG_OR_MID, EXPR_BEG),
	     check_modep(rbtok("/=/", :tREGEXP,  "/=/"), BEG_AND_MID),
	     set_modep(rbtok("/", "/"), EXPR_BEG))

    # Currently unhandled:
    #assert_tok(" /a/", :tREGEXP, "/a/", :prest => EXPR_ARG)
    #assert_tok("/a/", "/", "/", :prest => EXPR_ARG, :postst => EXPR_BEG)
    #assert_tok(" / ", "/", "/", :prest => EXPR_ARG, :postst => EXPR_BEG)

    LeftBrace  = choice(modep(rbtok("{", :tLBRACE), NOT_END_OR_ARG, EXPR_BEG),
			modep(rbtok("{", "{"),      END_AND_ARG,    EXPR_BEG))

    RightBrace = set_modep(rbtok("}", "}"), EXPR_END)

    # Newlines are treated differently depending on the mode. The newline
    # parser is used to update the position object but you can't see it
    # here because its defined in std_library's Newline.
    NewlineSkip = check_modep(Newline, 
			      [EXPR_BEG, EXPR_FNAME, EXPR_DOT])
    NewlineMatch = 
      rbtok(modep(Newline, 
		  EXPR_STATES - [EXPR_BEG, EXPR_FNAME, EXPR_DOT], EXPR_BEG),
	    "\n")

    LeftBracket  = 
      choice(check_modep(rbtok("[", "["), FNAME_AND_DOT),
	     modep(rbtok("[", :tLBRACK), BEG_AND_MID, EXPR_BEG),
	     modep(wsp(rbtok("[", :tLBRACK)), EXPR_ARG, EXPR_BEG),
	     set_modep(rbtok("[", "["), EXPR_BEG))
	     
    RightBracket = set_modep(rbtok("]", "]"), EXPR_END)

    SingleDot = string(".").not_followed_by(Digit)

    Dot =
      choice( modep(rbtok("...", :tDOT3), EXPR_STATES, EXPR_BEG),
	      modep(rbtok("..", :tDOT2),  EXPR_STATES, EXPR_BEG),
	      modep(rbtok(SingleDot, "."),  EXPR_STATES, EXPR_DOT) )

    MaybeSign = maybe(choice("-", "+"))

    def TokenGrammar.ruby_num_literal(digitParser)
      plus(skip(mult("_")), plus(digitParser))
    end

    def TokenGrammar.int_literal(prefixParser, digitParser,
				 skipPrefix = true,
				 errorMsg = "numeric literal without digits")
      sub = [MaybeSign]
      if "" != prefixParser
	sub.push (skipPrefix ? skip(prefixParser) : prefixParser)
      end
      body = ruby_num_literal(digitParser).or_raises(errorMsg)
      sub.push body.not_followed_by("_", "trailing '_' in number")
      seq(*sub)
    end

    StdIntegerLiteral = int_literal( symbols("1".."9"), maybe(Digit), false )

    # Use maybe so that it handles a single zero
    OctalLiteral      = int_literal( "0",  maybe(OctalDigit) )

    HexPrefix         = seq("0", choice("x", "X"))
    HexLiteral        = int_literal(HexPrefix, HexDigit   )

    BinPrefix         = seq("0", choice("b", "B"))
    BinaryLiteral     = int_literal(BinPrefix, BinaryDigit)

    Integer = 
      set_modep(rbtok2(choice(HexLiteral.apply(HexToIntFunction.new),
			      BinaryLiteral.apply(BinaryToIntFunction.new),
			      OctalLiteral.apply(OctalToIntFunction.new),
			      StdIntegerLiteral.apply(ToIntFunction.new)),
		       IntegerLit),
		EXPR_END)

    def TokenGrammar.num_literal_spec(digitParser)
      seq(digitParser, mult(skip(mult("_")), plus(digitParser)))
    end

    Nums = num_literal_spec(Digit)
    FloatExponent = seq(choice("e", "E"), MaybeSign, ruby_num_literal(Digit))

    FloatLiteral = seq(MaybeSign,
		       choice(seq(Nums,
				  choice(seq(".", Nums, maybe(FloatExponent)),
					 FloatExponent)),
			      seq(".", Nums, maybe(FloatExponent))))

    Float = rbtok2(FloatLiteral.apply(ToFloatFunction.new), FloatLit)

    Colon = 
      choice(modep(rbtok("::", :tCOLON3), BEG_AND_MID, EXPR_BEG),
	     modep(rbtok(wsp("::"), :tCOLON3), EXPR_ARG, EXPR_BEG),
	     set_modep(rbtok("::", :tCOLON2), EXPR_DOT),
	     modep(rbtok(":", ":"), [EXPR_END, EXPR_ENDARG], EXPR_BEG),
	     set_modep(rbtok(seq(":", look_ahead(WhiteSpace)), ":"), EXPR_BEG),
	     set_modep(rbtok(":", :tSYMBEG), EXPR_FNAME))

    UpAlpha = symbols("A".."Z")
    VarChar = symbols("a".."z", "A".."Z", "0".."9", "_")
    DownAlpha = symbols("a".."z")
    Fid = rbtok(seq(DownAlpha, mult(VarChar), symbols("?", "!")),
		:tFID)

    NonNumVarChar = symbols("a".."z", "A".."Z", "_")
    Variable = seq(NonNumVarChar, mult(VarChar))
    InstanceVar = rbtok2(seq(skip("@"),  Variable), InstanceVar)
    ClassVar    = rbtok2(seq(skip("@@"), Variable), ClassVar)

    GlobalSpecialChar = symbols("*", %{$}, "?", "!", "@", "/", "\\", ";", ",",
				".", "=", ":", "<", ">", "\"")

    GlobalVariable = rbtok2(seq(skip(%{$}), 
				choice(Variable, GlobalSpecialChar)), 
			    GlobalVar, nil, EXPR_END)

    BackReference     = rbtok2(seq(skip(%{$}), choice("&", "`", "'", "+")), 
			       BackRef, nil, EXPR_END)  

    NthReference      = rbtok2(seq(skip(%{$}), 
				   plus(Digit)).apply(ToIntFunction.new),
			       NthRef, nil, EXPR_END)

    Comma     = rbtok(",", ",", nil, EXPR_BEG)

    SemiColon = rbtok(";", ";", nil, EXPR_BEG)

    ARef = choice(modep(rbtok("[]", :tAREF), EXPR_FNAME, EXPR_FNAME),
		  modep(rbtok("[]", :tAREF), EXPR_DOT, EXPR_DOT))

    ASet = choice(modep(rbtok("[]=", :tASET), EXPR_FNAME, EXPR_FNAME),
		  modep(rbtok("[]=", :tASET), EXPR_DOT, EXPR_DOT))

    LeftParen = 
      choice(modep(rbtok("(", :tLPAREN, :tLPAREN), BEG_AND_MID, EXPR_BEG),
	     modep(rbtok("(", "("), NOT_BEG_OR_MID, EXPR_BEG))

    RightParen = set_modep(rbtok(")", ")"), EXPR_END)

    Comment = rbtok(seq("#", mult(none_of("\n\r"))), :tComment)

    # Probably shouldn't be AlphaNum below, but use that for now cause
    # I don't know what it should be... CHECK
    BCBegin = seq(Bol, string("=begin").not_followed_by(AlphaNum), RestOfLine)
    BCEnd   = seq(Bol, string("=end").not_followed_by(AlphaNum), RestOfLine)
    BlockComment = rbtok(seq(skip(BCBegin), any_until(skip(BCEnd))),
			 :tBlockComment)

    # THINK Maybe faster in the long run if we don't implement reserved
    # words in this special way but simply add them as strings since the
    # optimizer will fix them later.
    # FIX We don't handle do in any special way. We need to add the cond_p
    # etc stuff both here and when matching parentheses and args.
    class MakeIdentConstantAndReserved < MakeRaccToken
      def initialize(reservedWords)
	@reserved = reservedWords
      end

      def call(aParseResult, aParseState = nil, aPosition = nil)
	normal_token, special_token, next_mode = @reserved[aParseResult]
	pos = position(aParseResult, aParseState, aPosition)
	if normal_token
	  if aParseState.mode == EXPR_BEG
	    aParseState.mode = next_mode
	    RaccToken.new(normal_token, aParseResult, pos)
	  else
	    if normal_token != special_token
	      aParseState.mode = EXPR_BEG
	    else
	      aParseState.mode = next_mode
	    end
	    RaccToken.new(special_token, aParseResult, pos)
	  end
	else
	  first_char = aParseResult[0]
	  if first_char >= ?A && first_char <= ?Z
	    make_token(Constant, aParseResult, pos)
	  else
	    make_token(Identifier, aParseResult, pos)
	  end
	end
      end

      def make_token(token, aParseResult, position)
	tok = token[aParseResult]
	tok.position = position
	tok
      end
    end
    
    IdConstResMaker = MakeIdentConstantAndReserved.new(Ruby::RESERVED_WORDS)
    NonNumIdChar = symbols("a".."z", "A".."Z", "_") 
    IdChar = symbols("a".."z", "A".."Z", "_", "0".."9") 
    IdConstRes = 
      seq(NonNumIdChar, mult(IdChar)).apply(IdConstResMaker)

    def TokenGrammar.delimited(open, close)
      not_open_or_close = none_of(open + close)
      recp = recursive(open, mult(mult(not_open_or_close), recurse), 
		       close)
      inner = mult(mult(not_open_or_close), recp)
      seq(skip(open), inner, skip(close))
    end

    NestedDelim = choice(delimited("{", "}"),
			 delimited("(", ")"),
			 delimited("<", ">"),
			 delimited("[", "]"))

    AnyDelim = seq(skip(capture(none_of("qQwrx"), :dstring_delim)), 
		   any_until(skip(ref(:dstring_delim))))

    DString = 
      rbtok(set_modep(choice(seq(skip("\""), 
				 mult(choice(none_of("\""), 
					     seq(look_back("\\"), "\""))),
				 skip("\"")),
			     seq(skip("%", maybe("Q")),
				 choice(NestedDelim, AnyDelim))),
		      EXPR_END),
	    :tDSTRING)

    String = 
      choice(rbtok(modep(seq(skip("'"), 
			     mult(choice(none_of("'"), 
					 seq(look_back("\\"), "'"))),
			     skip("'")), NOT_FNAME_OR_DOT, EXPR_END),
		   :tSTRING),
	     rbtok(set_modep(seq(skip("%q"), choice(NestedDelim, AnyDelim)),
			     EXPR_END), :tSTRING),
	     rbtok(modep("'", FNAME_AND_DOT, EXPR_END), "'"))

    Regexp = 
      rbtok(set_modep(choice(seq(skip("/"),
				 mult(choice(none_of("/"), 
					     seq(look_back("\\"), "/"))),
				 skip("/")),
			     seq(skip("%r"),
				 choice(delimited("{", "}"),
					delimited("<", ">")))),
		      EXPR_END),
	    :tREGEXP)

    # The parser for Ruby tokens
    Parser = 
      seq(skip(mult(choice(WhiteSpace, NewlineSkip))),
	  choice(Comment, BlockComment, NewlineMatch,
		 GlobalVariable, ClassVar, InstanceVar, 
		 Fid,
		 LeftBrace, RightBrace, Comma, SemiColon,
		 DString, String, Regexp,
		 Equality, Exclamation, QuestionMark,
		 Dot, Colon,
		 Plus, Minus, Star, Bar, Ampersand, Slash,
		 Float, Integer,
		 ASet, ARef, 
		 LeftParen, RightParen, LeftBracket, RightBracket,
		 LessThan, GreaterThan,
		 Caret, Tilde,
		 NthReference, BackReference,
		 IdConstRes))

    #========================================================================
  end

  class Lexer
    # A lexer class for using the moded parse state
    class ModedLexer < Rockit::Parse::Lexer
      def initialize(tokenParser, *streamSpecifyingObjects)
	@parser = tokenParser
	@pstate = get_state(streamSpecifyingObjects)
      end

      def get_state(streamSpecifyingObjects)
	if streamSpecifyingObjects.length == 1 &&
	    ModedState === streamSpecifyingObjects.first
	  streamSpecifyingObjects.first
	else
	  ModedState.new(*streamSpecifyingObjects) 
	end
      end

      def mode; @pstate.mode; end
      def mode=(newMode); @pstate.mode = newMode; end
    end

    def initialize(*streamSpecifyingObjects)
      @lexer = ModedLexer.new(TokenGrammar::Parser, *streamSpecifyingObjects)
      @command_state = true
      self.state = EXPR_BEG
    end

    def lex(addStates = false)
      tokens = Array.new
      state = self.state
      token = next_token
      while token.token_type != false
	tokens.push state if addStates
	tokens.push token
	state = self.state
	token = next_token
      end
      tokens
    end

    # Methods delegating to @lexer
    def state=(newState); @lexer.mode = newState; end
    def state; @lexer.mode; end
    def position; @lexer.position; end

    def next_token
      begin
	@old_command_state, @command_state = @command_state, false
	token = @lexer.next_token
	update_state(token)
	token
      rescue UnexpectedEosError => e
	return RaccToken.new(false, false, @lexer.position)
      end
    end

    # Matz uses a state machine to keep track of when statements start and
    # stop. I should probably do it from inside the token parsers above
    # but here its in one place.
    def update_state(token)
      if [";", "(", "\n"].include?(token.lexeme)
	@command_state = true
      elsif [:tIDENTIFIER, :tCONSTANT, 
	     :tGVAR, :tIVAR, :tCVAR].include?(token.token_type)
	if (state == EXPR_BEG || state == EXPR_DOT ||
	    state == EXPR_ARG || state == EXPR_CMDARG)
	  self.state = @old_command_state ? EXPR_CMDARG : EXPR_ARG
	else
	  self.state = EXPR_END
	end
      end
    end
  end
end
