require 'rockit/parse/parse_error'
require 'rockit/parse/result_mapping'
require 'rockit/parse/char_set'

module Rockit
  module Parse
    # Abstract stream
    class Stream
      def current; raise NotImplementedError; end
      def advance; raise NotImplementedError; end
      def [](index); raise NotImplementedError; end
      def position; @position; end
      def past_eos?; raise NotImplementedError; end

      def position=(newPosition)
	@position = newPosition
      end

      @@stream_classes = Hash.new

      def Stream.factory(*o)
	@@stream_classes[type_signature(o)].new(*o)
      end

      def Stream.type_signature(o)
	o.map {|e| e.type}
      end

      def Stream.add_stream_class(klass, *typeSignature)
	@@stream_classes[typeSignature] = klass
      end

      def extract_sequence(startposition, skipIndices); end
    end

    class StringStream < Stream
      Stream.add_stream_class(self, String)

      def initialize(aString)
	super()
	@str = aString
	@position = 0
      end

      def current; @str[@position]; end
      def advance; @position += 1; end
      def [](index); @str[@position + index]; end
      def past_eos?; @position >= @str.length; end

      def extract_sequence(startposition, skipIndices)
	if skipIndices.length == 0
	  @str[startposition...@position]
	else
	  str = ""
	  index = startposition
	  while skipIndices.first && index >= skipIndices.first.last
	    skipIndices.shift
	  end
	  while index < @position
	    if skipIndices.length > 0 && index == skipIndices.first.first
	      index = skipIndices.first.last
	      skipIndices.shift
	    else
	      str << @str[index]
	      index += 1
	    end
	  end
	  str
	end
      end
    end

    # Just to try the stream factory
    class FileStream < Stream
      Stream.add_stream_class(self, File, String)

      def initialize(aFile, aString)
	@filename = aString
	@file = aFile
      end
    end

    class Parser
      attr_accessor :sub_parsers
      attr_accessor :expected

      # Flag indicating whether this parser will always produce a result for
      # the result stack if it matches. The default behavior is to not be
      # result producing. An example of a result producing parser is the
      # ApplyParser below.
      attr_reader :result_producing

      def initialize(*parsers)
	@sub_parsers = parsers
	@expected = []
	@case_sensitive = true
	@result_producing = false
      end

      def new_parser_state(*streamSpecifyingObjects)
	if streamSpecifyingObjects.length == 1 and
	    ParseState === streamSpecifyingObjects.first
	  streamSpecifyingObjects.first
	else
	  OptimizedParseState.new(*streamSpecifyingObjects)
	end
      end

      def parse_stream(*streamSpecifyingObjects)
	@state = new_parser_state(*streamSpecifyingObjects)
	@start_position = @state.position
	@state.split
	begin
	  matched = match_and_commit(@state)
	  if matched
	    result = @state.result || ""
	  else
	    @state.back
	    error = analyze_error(@state)
	    raise error if error
	  end
	  result
	ensure
	  @state.clear
	end
      end

      def make_error(aState)
	if aState.past_eos?
	  UnexpectedEosError.new(self, aState)
	else
	  NoMatchError.new(self, aState)
	end
      end

      # We implement zero-cost error handling. This means that we first
      # assume there will be no errors so we don't check for them.
      # If there is a problem the state is set to the start point and
      # analyze_error is called. It will try to parse again but this time
      # it will return the error encountered. This way we can optimize the
      # first non-error checking version for speed.
      def analyze_error(aState)
	matched = match(aState)
	if matched
	  return nil
	else
	  return make_error(aState)
	end
      end

      def parse(aString)
	parse_stream(aString)
      end

      def expecting(aString)
	p = self.clone
	p.expected = [aString]
	p
      end

      def expected
	if @expected.length > 0
	  @expected
	else
	  default_expected
	end
      end

      def default_expected
	@sub_parsers.map {|p| p.expected}.flatten
      end

      attr_reader :case_sensitive

      def case_sensitive=(newValue)
	@case_sensitive = newValue
	@sub_parsers.each {|sp| sp.case_sensitive = newValue}
      end

      # Match and produce a new result but do not extract it
      def match_and_commit(aState = @state)
	aState.new_result
	res = match(aState)
	aState.commit_result if res
	res
      end

      def inspect(myself = nil)
	s = self.type.inspect.split("::").last
	s =~ /(.+)Parser/
	s = $1 if $1
	if @sub_parsers.length > 0
	  s += "(" + @sub_parsers.map{|p| p.inspect(myself)}.join(", ") + ")"
	end
	s
      end

      def first_set
	@first_set || (@first_set = calc_first_set())
      end

      def calc_first_set
	@sub_parsers.first.first_set
      end

      private

      # Match but add to existing result. Return true iff matched.
      def match(aState); raise NotImplementedError; end
    end

    class Position
      def initialize
	@pos, @line, @column = 0, 1, 1
      end

      def advance
	@column += 1
	@pos += 1
      end

      def <=>(o)
	if Numeric === o
	  @pos <=> o
	else
	  raise "Trying to compare a non-numerical object with a Position"
	end
      end

      def newline_matched
	@line += 1
      end

      def inspect
	"line = #{line}, column = #{@column}"
      end
    end

    class ParseState
      def initialize(*streamSpecifyingObjects)
	@stream = Stream.factory(*streamSpecifyingObjects)
	@attributes = Hash.new
	@splits = Array.new
	#@position = Position.new
      end

      # Start collecting a new result. If there are uncommitted results they
      # are discarded.
      def new_result; raise NotImplementedError; end

      # Commit result => finish the current result collection and move result
      # to result stack. Prepare for new_result.
      def commit_result; raise NotImplementedError; end

      # Accessor for the result stack. Without parameters it returns the last
      # result. If an index is given it must be negative with -1 being the latest
      # result and -2 the next latest. This function never modifies the result
      # stack.
      def result(index = -1); raise NotImplementedError; end

      # Pop the latest result from result stack
      def pop_result; raise NotImplementedError; end

      # Pop the latest results from result stack
      def pop_results(count); raise NotImplementedError; end

      # Push to result stack
      def push_result(aResult); raise NotImplementedError; end

      def extract_result(startposition)
	@stream.extract_sequence(startposition, [])
      end


      # Skip symbols matched in code we yield to. Must handle nested skips.
      def skip; raise NotImplementedError; end

      # Match the current symbol and advance the cursor. Since this is more 
      # common than skip we use a reversed scheme where we only save 
      # information about the skips.
      def advance
	#@position.advance
	@stream.advance
      end

      def position; @stream.position; end
      def position=(newPos)
	@stream.position = newPos
      end

      # Called after an advance over the newline symbol(s) so should *NOT*
      # update the stream position.
      def newline_matched
	# Not yet implemented
      end

      def past_eos?; @stream.past_eos?; end
      def current; @stream.current; end
      def [](index); @stream[index]; end

      def set_attribute(key, value)
	@attributes[key] = value
      end

      def get_attribute(key)
	@attributes[key]
      end

      def split
	@splits.push position
      end

      def back
	@stream.position = @splits.last
      end

      def clear
	@splits.pop
      end
    end

    class OptimizedParseState < ParseState
      def initialize(*args)
	super(*args)
	@results = Array.new
	@skips = Array.new	   # A skip is the start and stop positions of a skip
      end

      def new_result
	@start = position
      end

      def uncommitted?
	@start != position
      end

      def commit_result
	if uncommitted?
	  @results.push @stream.extract_sequence(@start, @skips)
	  new_result
	end
      end

      def result(index = -1)
	@results[index]
      end
      
      def pop_result
	@results.pop
      end

      def pop_results(count)
	ary = Array.new(count)
	count.times {|i| ary[count-i-1] = @results.pop}
	ary
      end

      def push_result(aResult)
	@results.push aResult
      end

      def skip
	start_skipping
	begin
	  res = yield
	ensure
	  stop_skipping(res)
	end
	res
      end

      def back
	@stream.position = @splits.last
	delete_skips_after_position(@splits.last)
      end

      private

      def start_skipping
	if @skip_start
	  @skip_nest += 1
	else
	  @skip_start = position
	  @skip_nest = 1
	end
      end

      def stop_skipping(result)
	@skip_nest -= 1
	if @skip_nest == 0
	  if result && position > @skip_start
	    @skips.push [@skip_start, position]
	  end 
	  @skip_start = nil
	end
      end

      def delete_skips_after_position(position)
	while @skips.last && @skips.last.first >= position
	  @skips.pop
	end
      end
    end

    # Matches a set of symbols
    class SymbolsParser < Parser
      attr_reader :symbols

      def initialize(*symbols)
	super()
	@symbols = symbols
      end

      def merge(aParser)
	unless aParser.type == self.type
	  raise ArgumentError, "#{self.type}'s can only be merged with #{self.type}'s"
	end
	merged = self.type.new(*@symbols.concat(aParser.symbols))
	merged.expected = self.expected + aParser.expected
	merged
      end

      def default_expected
	@symbols
      end

      def match(aState)
	return nil if aState.past_eos?
	if match_symbol(aState.current)
	  aState.advance
	  return true
	end
      end

      def match_symbol(symbol)
	@symbols.include?(symbol)
      end

      def inspect(myself = nil)
	"Symbols(" + symbols.inspect[1...-1] + ")"
      end
    end

    # A fast way to do symbol matching when the symbols are chars is to
    # do it with an Array for all the 256 possible char values with the
    # string corresponding to the matched string.
    class OneOfParser < SymbolsParser
      def initialize(*theSymbols)
	if theSymbols.length == 1 && String === theSymbols.first &&
	    theSymbols.first.length >= 1
	  theSymbols = symbols_from_string(theSymbols.first)
	end
	super(*theSymbols)
	@fast_array = make_fast_array(theSymbols)
      end

      def symbols_from_string(aString)
	aString.split("").map {|s| s[0]}
      end
      
      def make_fast_array(symbols)
	a = Array.new
	symbols.each {|e| a[e] = true}
	a
      end

      def match_symbol(symbol)
	@fast_array[symbol]
      end

      def case_sensitive=(newValue)
	return if case_sensitive == newValue
	super(newValue)
	unless @chars_with_inverse_case
	  # This is the first time so we must calculate the chars with inverse
	  # case that should be added and add them
	  @chars_with_inverse_case = inverse_case(@symbols)
	  @symbols += @chars_with_inverse_case
	else
	  # This is not the first time so @chars_with_inverse_case already
	  # calculated
	  if newValue
	    @symbols -= @chars_with_inverse_case
	  else
	    @symbols += @chars_with_inverse_case
	  end
	end
	@fast_array = make_fast_array(@symbols)
      end

      def inverse_case(chars)
	ary = Array.new
	chars.each do |char| 
	  as_string = char.chr
	  t = as_string.upcase
	  ary.push(t == as_string ? as_string.downcase[0] : as_string.upcase[0])
	end
	ary
      end

      def inspect(myself = nil)
	if symbols.length == 1
	  "'#{symbols[0].chr}'"
	else
	  "OneOfParser('" + symbols.map {|c| c.chr}.join + "')"
	end
      end

      def calc_first_set
	CharSet.new(symbols.map {|c| c.chr}.join)
      end
    end

    def one_of(aString)
      OneOfParser.new(aString)
    end

    def symbols(*objects)
      objects = unpack_ranges(objects)
      objects = strings_to_chars(objects)
      if all_characters?(objects)
	OneOfParser.new(*objects)
      else
	SymbolsParser.new(*objects)
      end
    end
    def symbol(object); symbols(object); end

    def unpack_ranges(objects)
      (objects.map {|e| Range === e ? e.to_a : [e]}).flatten
    end

    def strings_to_chars(objects)
      objects.map {|e| (String === e && e.length == 1) ? e[0] : e}
    end

    def all_characters?(anArray)
      anArray.all? {|e| Fixnum === e && 0 <= e && e < 256}
    end

    class CompoundParser < Parser; end

=begin
    # There are problems with using this flattening to simplify since
    # the user might hold onto sub_parsers in order to know when to stop
    # when inserting separators.
    #
    # The performance gain when using this seems fairly modest (about 3%
    # when tried on the utest_lexer in examples/ruby) and should be even
    # less noticeable when rewriting the parser for performance using factoring
    # etc.
    def flatten(klass, parsers)
      if parsers.any? {|p| self.type === p}
	flattened_parsers = []
	parsers.each do |p|
	  if klass === p
	    flattened_parsers.concat p.sub_parsers
	  else
	    flattened_parsers << p
	  end
	end 
	parsers = flattened_parsers
      end
      parsers
    end
=end

    class ChoiceParser < CompoundParser
      def initialize(*parsers)
	super(*parsers)
	@result_producing = sub_parsers.any? {|p| p.result_producing}
      end
      
      def match(aState)
	aState.split
	begin
	  @sub_parsers.each do |parser|
	    return true if parser.match(aState)
	    aState.back
	  end
	  false
	ensure
	  aState.clear
	end
      end

      def analyze_error(aState)
	error = nil
	aState.split
	begin
	  @sub_parsers.each do |parser|
	    e = parser.analyze_error(aState)
	    error = (error ? error.merge(e) : e) if e
	    aState.back
	  end
	  error
	ensure
	  aState.clear
	end
      end

      def calc_first_set
	@sub_parsers.inject(CharSet.new) {|set, p| set.merge(p.first_set)}
      end
    end

    def choice(*parsersOrSymbols)
      make_compound_parser(ChoiceParser, *parsersOrSymbols)
    end

    def make_compound_parser(klass, *parsersOrSymbols)
      parsers = make_parsers(parsersOrSymbols)
      if parsers.length > 1
	klass.new(*parsers)
      else
	parsers.first
      end
    end

    def make_parsers(parsersOrSymbols)
      parsersOrSymbols.map {|o| make_parser(o)}
    end

    def make_parser(o)
      if Parser === o
	o 
      elsif String === o and o.length > 1
	string(o)
      else
	symbols(o)
      end
    end

    class SequenceParser < CompoundParser
      def initialize(*matchers)
	super(*matchers)
      end

      def match(aState)
	@sub_parsers.each do |parser|
	  return nil unless parser.match(aState)
	end
	true
      end

      def analyze_error(aState)
	@sub_parsers.each do |parser|
	  e = parser.analyze_error(aState)
	  return e if e
	end
	nil
      end

      def default_expected
	index, expected_strings = 0, [nil]
	while index < @sub_parsers.length &&
	    expected_strings.include?(nil) # Indicates that it matches epsilon
	  expected_strings.compact!
	  expected_strings.concat @sub_parsers[index].expected
	  index += 1
	end
	expected_strings
      end

      def calc_first_set
	set = @sub_parsers.inject(CharSet.new) do |set, p| 
	  set.merge(p.first_set)
	  derives_epsilon = set.include_epsilon?
	  set.remove_epsilon
	  return set unless derives_epsilon
	  set
	end
	set.add_epsilon # If we get here epsilon was in all first sets
	set
      end      
    end

    def seq(*parsersOrSymbols)
      make_compound_parser(SequenceParser, *parsersOrSymbols)
    end

    class RepeatParser < Parser
      def initialize(matcher, min = 0, max = nil)
	super(matcher)
	@min, @max = min, max
	@result_producing = matcher.result_producing
	if @result_producing
	  class <<self; alias match match_and_push_result; end
	end
      end

      def higher_than_equal_max(value)
	@max and value >= @max
      end

      def default_expected
	e = @sub_parsers.first.expected
	e.push nil if @min == 0 # Indicate that it matches empty string
	e
      end

      def match_and_push_result(aState)
	parser, count, last_pos = @sub_parsers.first, 0, -1
	aState.new_result
	while parser.match(aState) && last_pos != aState.position
	  aState.commit_result
	  last_pos = aState.position
	  count += 1
	  break if higher_than_equal_max(count)
	end
	aState.position = last_pos unless last_pos == -1
	aState.commit_result
	if count >= @min
	  aState.push_result(get_results(count, aState))
	  true
	else
	  aState.pop_results(count) # Reset result stack
	  nil
	end
      end

      def get_results(count, aState)
	results = aState.pop_results(count)
	results.map {|s| (s=="" ? nil : s)}.compact
      end
	
      def match(aState)
	parser, count, last_pos = @sub_parsers.first, 0, -1
	while parser.match(aState) && last_pos != aState.position
	  last_pos = aState.position
	  count += 1
	  break if higher_than_equal_max(count)
	end
	aState.position = last_pos unless last_pos == -1
	(count >= @min) ? true : nil
      end

      def inspect(myself = nil)
	case [@min, @max]
	when [0,1]
	  "Maybe(#{@sub_parsers.first.inspect(myself)})"
	when [0,nil]
	  "Mult(#{@sub_parsers.first.inspect(myself)})"
	when [1,nil]
	  "Plus(#{@sub_parsers.first.inspect(myself)})"
	else
	  "Repeat(#{@sub_parsers.first.inspect(myself)}, #{@min}" +
	    (@max ? ", #{@max}" : "") +
	    ")"
	end
      end

      def calc_first_set
	set = super
	set.add_epsilon if @min == 0
	set
      end
    end

    def repeat(*matchers)
      if Numeric === matchers[-2]
	max = matchers.pop
	min = matchers.pop
      elsif Numeric === matchers[-1]
	max, min = nil, matchers.pop
      else
	max, min = nil, 0
      end
      RepeatParser.new(seq(*matchers), min, max)
    end

    def plus(*matchers)
      RepeatParser.new(seq(*matchers), 1)
    end

    def mult(*matchers)
      RepeatParser.new(seq(*matchers), 0)
    end

    def maybe(*matchers)
      RepeatParser.new(seq(*matchers), 0, 1)
    end

    def string(aString)
      seq(*aString.split(""))
    end
    module_function :string

    class Parser
      def apply(anObject)
	result_mapping_function = make_result_mapping_function(anObject)
	ApplyParser.new(self, result_mapping_function)
      end

      private

      def make_result_mapping_function(anObject)
	if String === anObject
	  ProcFunction.new(make_proc_from_string(anObject))
	elsif Proc === anObject
	  ProcFunction.new(anObject)
	elsif ResultMappingFunction === anObject
	  anObject
	else
	  raise ArgumentError, "Argument is not a String, Proc or ResultMappingFunction but #{anObject.type.inspect}"
	end
      end

      def make_proc_from_string(aString)
	aString = "proc" + aString unless aString =~ /^proc/
	begin
	  p = eval(aString)
	rescue Exception
	end
	if Proc === p
	  p
	else
	  raise ArgumentError, "The string '#{aString}' cannot be converted to a Proc object"
	end
      end
    end

    class ApplyParser < Parser
      def initialize(aParser, aResultMappingFunction)
	super(aParser)
	@function = aResultMappingFunction
      end

      def result_producing
	true
      end

      def match_and_commit(aState)
	# They are the same since ApplyParser will always produce a result
	match(aState)
      end

      def match(aState)
	@position = aState.position
	aState.new_result
	res = @sub_parsers.first.match(aState)
	if res
	  aState.commit_result
	  res = @function.call(aState.pop_result, aState, @position)
	  aState.push_result(res)
	end
	res
      end
    end

    class NotParser < Parser
      attr_accessor :reset

      def initialize(aParser, advance = true, aString = nil)
	super(aParser)
	@reset, @advance, @message = true, advance, aString
      end

      def match(aState)
	if match_sub_parser(aState)
	  nil
	else
	  aState.advance if @advance
	  true
	end
      end

      def match_sub_parser(aState)
	aState.split
	res = @sub_parsers.first.match(aState)
	aState.back if @reset
	aState.clear
	res
      end

      def analyze_error(aState)
	if match_sub_parser(aState)
	  if @message
	    CustomParseError.new(@message, aState)
	  else
	    UnexpectedError.new(aState)
	  end
	end
      end

      def default_expected
	[]
      end

      def calc_first_set
	super.invert
      end
    end

    def notp(o, advance = true, aString = nil)
      NotParser.new(make_parser(o), advance, aString)
    end

    class Parser
      def not_followed_by(aParser, aString = nil)
	seq(self, notp(aParser, false, aString))
      end
    end

    class SkipParser < Parser
      def match(aState)
	aState.skip {@sub_parsers.first.match(aState)}
      end

      def analyze_error(aState)
	@sub_parsers.first.analyze_error(aState)
      end
    end

    def skip(*os)
      SkipParser.new(seq(*make_parsers(os)))
    end

    class FailParser < Parser
      def initialize(aString, aParser = nil)
	super(aParser) # Parser not used in this class though!
	@message = aString
      end
      
      def match(aState)
	raise error(aState)
      end
      
      def error(aState)
	CustomParseError.new(@message, aState)
      end
    end

    class FailIfMatchParser < FailParser
      def initialize(aString, aParser, reset = true)
	super(aString, aParser)
	@reset = reset
      end

      def match(aState)
	begin
	  aState.split
	  result = @sub_parsers.first.match(aState)
	  aState.back if result && @reset
	ensure
	  aState.clear
	end
	raise error(aState) if fail?(result)
	true
      end

      def fail?(result)
	result
      end
    end

    class FailUnlessMatchParser < FailIfMatchParser
      def fail?(result)
	not result
      end
    end

    def failp(aString)
      FailParser.new(aString)
    end

    class Parser
      def raises(aString)
	FailIfMatchParser.new(aString, self)
      end

      def or_raises(aString)
	FailUnlessMatchParser.new(aString, self, false)
      end
    end

    class LookBackParser < Parser
      def initialize(aParser, matchStartOfString = false)
	unless valid_lookback_parser?(aParser)
	  raise "Only strings and symbol parsers allowed in lookback"
	end
	super(aParser)
	@length = 
	  (SequenceParser === aParser) ? aParser.sub_parsers.length : 1
	@match_start_of_string = matchStartOfString
      end

      # TODO: generalize so that all parsers with a fix length can be used
      def valid_lookback_parser?(aParser)
	if SymbolsParser === aParser
	  true
	elsif SequenceParser === aParser
	  aParser.sub_parsers.all? {|sp| SymbolsParser === sp}
	end
      end

      def match(aState)
	if @length > aState.position
	  return @match_start_of_string if aState.position == 0
	  nil
	else
	  position = aState.position
	  aState.position -= @length
	  res = @sub_parsers.first.match(aState)
	  aState.position = position
	  res
	end
      end

      def calc_first_set
	CharSet.new("", true)
      end
    end

    def look_back(o, matchStartOfString = false)
      LookBackParser.new(make_parser(o), matchStartOfString)
    end

    class NoneOfParser < OneOfParser
      def initialize(aString)
	super(*symbols_from_string(aString))
      end

      def match_symbol(symbol)
	not @fast_array[symbol]
      end

      def calc_first_set
	super.invert
      end
    end

    def none_of(aString)
      NoneOfParser.new(aString)
    end

    # Delete this? Legacy???
    class NewlineParser < Parser
      def initialize(aParser)
	super(aParser)
      end

      def match(aState)
	res = @sub_parsers.first.match(aState)
	aState.newline_matched if res
	res
      end
    end

    def newline(o)
      NewlineParser.new(make_parser(o))
    end

    class AnyUntilParser < Parser
      def initialize(aParser, consume = true)
	super(aParser)
	@consume = consume
	@search_set = aParser.first_set
	if @search_set.include_epsilon?
	  raiser ArgumentError, "Parser can match ''; not allowed"
	end
      end

      def match(aState)
	parser, match = @sub_parsers.first, nil
	until aState.past_eos? || match 
	  while not @search_set.include_char?(aState.current)
	    aState.advance
	    break if aState.past_eos?
	  end
	  aState.split unless @consume
	  match = parser.match(aState)
	  aState.back if !@consume and match
	end
	match
      end

      def calc_first_set
	super.invert
      end
    end

    def any_until(o, consume = true)
      AnyUntilParser.new(make_parser(o), consume)
    end

    class LookAheadParser < Parser
      def initialize(aParser, matchStartOfString = false)
	super(aParser)
      end

      def match(aState)
	aState.split
	begin
	  @sub_parsers.first.match(aState)
	ensure
	  aState.back
	  aState.clear
	end
      end

      def calc_first_set
	CharSet.new("", true)
      end
    end

    def look_ahead(o)
      LookAheadParser.new(make_parser(o))
    end

    class SequenceParser
      def lift(anIndex)
	if anIndex >= @sub_parsers.length
	  raise "No parsers matching index #{anIndex}"
	end
	a = Array.new(anIndex+1).map {:_} # Mark all to be skipped
	a[anIndex] = :^                   # Mark the requested result to be returned
	TermProducingSequenceParser.new(@sub_parsers, a)
      end
    end

    class TermProducingSequenceParser < SequenceParser
      def initialize(parsers, resultSpecification, aTerm = nil)
	super(*parsers)
	@result_specs = resultSpecification
	@term = aTerm
      end

      def result_producing
	true
      end

      def match(aState)
	num_results = 0
	@sub_parsers.each_with_index do |parser, i|
	  result_spec = @result_specs[i]
	  aState.new_result if result_spec && result_spec != :_
	  unless parser.match(aState)
	    return clean_up_and_exit(aState, num_results)
	  end
	  if result_spec && result_spec != :_
	    aState.commit_result
	    num_results += 1
	  end
	end
	produce_result(aState, num_results)
      end

      def clean_up_and_exit(aState, numResults)
	aState.pop_results(numResults)
	nil
      end

      def produce_result(aState, numResults)
	results = aState.pop_results(numResults)
	if @term
	  @term.new(*results)
	else
	  if results.length > 1
	    aState.push_result results
	  else
	    aState.push_result results.first
	  end
	end
      end
    end

    class CaptureParser < Parser
      def initialize(aParser, aName)
	super(aParser)
	@name = aName
      end

      def match(aState)
	position = aState.position
	res = @sub_parsers.first.match(aState)
	if res
	  result = aState.extract_result(position)
	  aState.set_attribute(@name, result)
	end
	res
      end
    end

    def capture(aParser, aName)
      CaptureParser.new(make_parser(aParser), aName)
    end

    class ReferencingParser < Parser
      def initialize(aName)
	super()
	@name = aName
      end

      def match(aState)
	string(aState.get_attribute(@name)).match(aState)
      end
    end

    def ref(aName)
      ReferencingParser.new(aName)
    end

    class SucceedParser < Parser
      def initialize(value = nil)
	super()
	@value = value
      end

      def match(aState)
	aState.push_result(@value) if @value
	true
      end
    end

    def succeed(result = nil)
      SucceedParser.new(result)
    end

    class RecursiveParser < SequenceParser
      def initialize(*parsers)
	super(*insert_self(parsers))
      end

      def inspect(myself = nil)
	if myself
	  "recurse"
	else 
	  "recursive(#{@sub_parsers.map{|m| m.inspect(myself || self)}.join(', ')})"
	end
      end

      protected

      def insert_self(parsers, visited = {self => true})
	parsers.map! do |parser|
	  if Recurse === parser
	    self
	  elsif Parser === parser
	    unless visited[parser.id]
	      visited[parser.id] = true
	      insert_self(parser.sub_parsers, visited)
	    end
	    parser
	  else
	    parser
	  end
	end
      end
    end

    def recursive(*parsers)
      RecursiveParser.new(*make_parsers(parsers))
    end

    class Recurse < Parser; end

    def recurse
      Recurse.new
    end

    # Same as SequenceParser but return results of each subparser in array.
    class SeparatingSequenceParser < SequenceParser
      def match(aState)
	count = 0
	@sub_parsers.each do |parser|
	  aState.new_result
	  return nil unless parser.match(aState)
	  aState.commit_result
	  count += 1
	end
	aState.push_result(get_results(count, aState))
	true
      end

      def get_results(count, aState)
	results = aState.pop_results(count)
	results.map {|s| (s=="" ? nil : s)}.compact
      end
    end

    def sepseq(*parsersOrSymbols)
      make_compound_parser(SeparatingSequenceParser, *parsersOrSymbols)
    end    

    def separated_results(aParser)
      if SequenceParser === aParser
	SeparatingSequenceParser.new(*aParser.sub_parsers)
      else
	aParser
      end
    end

    class BuildTerm < ResultMappingFunction
      def initialize(aTermClass)
	@term_class = aTermClass
      end

      def call(aParseResult, aParseState = nil, aPosition = nil)
	if Array === aParseResult
	  @term_class[*aParseResult]
	else
	  @term_class[aParseResult]
	end
      end
    end

    class Parser
      def build(aTermClass)
	separated_results(self).apply(BuildTerm.new(aTermClass))
      end

      def |(other)
	choice(self, other)
      end

      def &(other)
	seq(self, other)
      end
    end
  end # module Parse
end # module Rockit
