module Rockit
  module Parse
    # Class representing functions that are applied to parse results.
    # We use a class since we want to encourage people to make them dumpable
    # to Ruby and C. This way we can dump parsers to code in different 
    # languages for speed. This abstract class does not require that you add
    # methods for dumping to source code. If you don't there will be errors
    # if you try anyway.
    # THINK: should they be singletons?
    class ResultMappingFunction
      def to_ruby_code; raise NotImplementedError; end
      def to_c_code; raise NotImplementedError; end

      def call(aParseResult, aParseState = nil, aPosition = nil)
	aParseResult # Identity function
      end
    end

    # Result mapper for converting float strings to floats. We cheat and use
    # Ruby...
    class ToFloatFunction < ResultMappingFunction
      def call(aParseResult, aParseState = nil, aPosition = nil)
	begin
	  eval(aParseResult)
	rescue Exception
	  raise ArgumentError, "Could not convert '#{aParseResult}' to float"
	end
      end
    end

    # Result mapper for converting (base 10) integer strings to integers
    class ToIntFunction < ResultMappingFunction
      def initialize(base = 10)
	super()
	@mapping, @base = calc_mapping, base
      end

      def calc_mapping
	a = Array.new
	index = 0
	(?0..?9).each {|c| a[c] = index; index += 1}
	a
      end

      def call(aParseResult, aParseState = nil, aPosition = nil)
	res = 0
	if ?- == aParseResult[0]
	  sign = -1
	  aParseResult = aParseResult[1..-1]
	elsif ?+ == aParseResult[0]
	  sign = 1
	  aParseResult = aParseResult[1..-1]
	else
	  sign = 1
	end
	aParseResult.each_byte do |char|
	  val = @mapping[char]
	  raise ArgumentError, "The string '#{aParseResult}' could not be converted to an integer" unless val
	  res = res * @base + val
	end
	res * sign
      end
    end

    # Result mapper for converting (base 16) hexadecimal integer strings to 
    # integers
    class HexToIntFunction < ToIntFunction
      def initialize
	super(16)
      end

      def calc_mapping
	a = Array.new
	index = 0
	(?0..?9).each {|c| a[c] = index; index += 1}
	("a".."f").each {|s| a[s[0]] = a[s.upcase[0]] = index; index += 1}
	a
      end
    end

    # Result mapper for converting (base 8) octal integer strings to 
    # integers
    class OctalToIntFunction < ToIntFunction
      def initialize
	super(8)
      end

      def calc_mapping
	a = Array.new
	index = 0
	(?0..?7).each {|c| a[c] = index; index += 1}
	a
      end
    end

    # Result mapper for converting (base 2) hexadecimal integer strings to 
    # integers
    class BinaryToIntFunction < ToIntFunction
      def initialize
	super(2)
      end

      def calc_mapping
	a = Array.new
	index = 0
	(?0..?1).each {|c| a[c] = index; index += 1}
	a
      end
    end

    # A function is simply a wrapper that allow you to use Proc objects
    # as RMF's.
    class ProcFunction < ResultMappingFunction
      def initialize(block)
	super()
	@block = block
      end

      def call(aParseResult, aParseState = nil, aPosition = nil)
	case @block.arity
	when 3
	  @block.call(aParseResult, aParseState, aPosition)
	when 2
	  @block.call(aParseResult, aParseState)
	else
	  @block.call(aParseResult)
	end
      end
    end

    # Convert a character to Int
    class CharToIntFunction < ResultMappingFunction
      def call(aParseResult, aParseState = nil, aPosition = nil)
	aParseResult[0] & 0xff
      end
    end    
    
  end # module Parse
end # module Rockit

