# Standard Library of Parsers
# $Id: std_library.rb,v 1.6 2002/01/06 21:19:02 feldt Exp $
#
# Copyright (c) 2001 Robert Feldt, feldt@ce.chalmers.se
#
require 'rockit/parse/combinator_parsers'
include Rockit::Parse
require 'rockit/term/term'
include Rockit

module Rockit
  module Parse
    Digit       = symbols("0".."9").expecting("a digit")
    HexDigit    = symbols("0".."9", "a".."f", "A".."F").expecting("a hexdigit")
    OctalDigit  = symbols("0".."7").expecting("an octal digit")
    BinaryDigit = symbols("0".."1").expecting("a binary digit")

    Upper       = symbols("A".."Z").expecting("an uppercase letter")
    Lower       = symbols("a".."z").expecting("a lowercase letter")
    Letter      = symbols("a".."z", "A".."Z").expecting("a letter")
    AlphaNum    = choice(Letter, Digit).expecting("an alphanumerical char")

    Space       = one_of(" \t\n\r\v\f")
    Spaces      = mult(Space)
    Tab         = symbols("\t").expecting("a tab character")

    AnyChar     = symbols(*(0...256).map{|c| c.chr}).expecting("a char")

    NewlineUnix = newline(symbols("\n").expecting("a newline"))
    NewlineWin  = newline(string("\r\n").expecting("a newline"))
    Newline     = choice(NewlineUnix, NewlineWin)

    RestOfLine  = seq(mult(none_of("\n\r")), Newline)

    # Beginning of a line. This is just an "anchor" so returns "".
    Bol = look_back("\n", true)

    # --------------------------------------------------------------------
    # Convenience parser generators
    # --------------------------------------------------------------------
    def nest(open, close)
      recursive(open,
		any_until(choice(open, close), false), 
		maybe(seq(recurse, any_until(close, false))),
		close)
    end

    # Derived parsers for convenience or speed

    # Get lines until one match
    class LinesUntilParser < Parser
      def initialize(aParser)
	super
      end

      def match(aState)
	aState.split
	begin
	  while not aState.past_eos?
	    res = @sub_parsers.first.match(aState)
	    break if res
	    # Search for line end. Will always be "\n" since both Unix ("\n") 
	    # and Windows ("\r\n") convention ends with "\n".
	    while (not (eos = aState.past_eos?)) && ?\n != aState.current
	      aState.advance
	    end
	    break if eos
	    aState.advance # Skip \n
	  end
	  aState.back unless res
	  res
	ensure
	  aState.clear
	end
      end
    end

    def lines_until(*parsers)
      LinesUntilParser.new(seq(*parsers))
    end

    class Parser
      def &(other)
	seq(self, other)
      end

      def |(other)
	choice(self, other)
      end
    end

    # Create a Term subclass and (optionally) a skip specification. Can
    # be called in multiple ways:
    #  term() - without arguments this returns Term meaning that we don't
    #           care to name the term or its children but will access them
    #           via []
    #  term(:Name, *childrens) - create a Term subclass named Name.
    #                            If the childrens spec contains :_ the 
    #                            corresponding result should be skipped.
    #                            Since we don't want to perform the skipping
    #                            when the term is created but in the
    #                            corresponding sub-parsers we return a 
    #                            TermWithSkips and let the 
    #                            build or ">" functions update the parsers.
    def term(name = nil, *childrens)
      if name
	if childrens.any? {|cn| cn == :_}
	  skip_indices = []
	  childrens.each_with_index do |c, i|
	    if c == :_
	      childrens[i] = nil
	      skip_indices.push i
	    end
	  end
	  TermWithSkips.new_sub_class_with_skips(skip_indices, name, 
						 *childrens.compact)
	else
	  Term.new_sub_class(name, *childrens)
	end
      else
	Term
      end
    end

    # Shorthand for creating a term creating parser
    class Parser
      def ^(term)
	if term.ancestors.include?(TermWithSkips)
	  self.clone.skip_subparsers(term.___skip_indices___).build(term)
	else
	  self.build(term)
	end
      end

      protected

      def skip_subparsers(indices)
	@sub_parsers.each_with_index do |p, i| 
	  @sub_parsers[i] = skip(p) if indices.include?(i)
	end
	self
      end
    end

    # Short hand for choice when we need to stack parser on top of each other
    # We can't use "|" since it has the same precedence as "^". Maybe there is
    # a combination of three operators that are better and still assign 
    # precedence in the order seq, build, choice. But I haven't found them
    # yet.
    class Parser
      def >(other)
	choice(self, other)
      end
    end

    # Add a separator that when set on a result_producing parser will make
    # it allow the things matched by the separator parser before and between
    # its subparsers. Since it only applied to result_producing parsers it
    # will not mess up token/lexer parsers.
    class Parser
      attr_reader :separator

      def set_separator(parser, tokenParsers = [])
	set_separator_recurse(make_parser(parser), tokenParsers)
      end

      def set_separator_recurse(parser, tokenParsers = [], visited = {})
	return if tokenParsers.include?(self) || visited[self.id]
	visited[self.id] = true
	@separator = parser
	@sub_parsers.each {|sp| 
	  sp.set_separator_recurse(@separator, tokenParsers, visited)
	}
	# Should really be the common base class of Seq and ChoiceParsers
	# when there is one...
	if SequenceParser === self || ChoiceParser === self
	  @sub_parsers = interleave_parser(@separator, @sub_parsers)
	end
      end

      private

      def interleave_parser(parser, parsers)
	new_parsers = Array.new
	parsers.each {|p| new_parsers << parser << p}
	new_parsers
      end
    end
  end # Parse

  # A special Term class for Term's with skip indices
  class TermWithSkips < Term
    extend PartialEvalSubClassFactory
    def self.new_sub_class_with_skips(skipIndices, name, *childrenNames)
      cls = new_sub_class(name, *childrenNames)
      params = "@@skips" + rand(1e10).inspect # Make it hard to guess
      cls.class_eval %{
       #{params} = []
       def self.___skip_indices___; #{params}[0]; end
      }
      (cls.class_eval(params) << skipIndices).freeze
      cls
    end
  end

  # It is recommended that grammars are written in their own module.
  # By extending with this Grammar module they get some nice class methods.
  module Grammar
    def language(string = nil)
      string ? (@language = string) : @language
    end
    
    def version(string = nil)
      string ? (@version = string) : @version
    end
    
    def start(parser = nil)
      return @start_parser if nil == parser
      raise ArgumentError, "Must be a Parser" unless Parser === parser
      @start = parser
    end
    
    def parse(string)
      @start.parse(string)
    end
  end # Grammar
end

class String
  def to_parser
    Rockit::Parse::string(self)
  end

  def |(o)
    choice(self.to_parser, o)
  end

  def &(o)
    seq(self.to_parser, o)
  end
end
