require 'rubyunit'

require 'rockit/parse/combinator_parsers'
require 'rockit/parse/result_mapping'
require 'rockit/parse/std_library'
include Rockit::Parse

require 'rockit/term/term'
include Rockit

class CombinatorParsersTest < RUNIT::TestCase
  def assert_parse(string, parser, expected = string)
    begin
      assert_equals(expected, parser.parse(string))
    rescue Exception => e
      assert(false, "Parser #{parser.inspect} does no match '#{string}'")
    end
  end

  def test_01_symbol
    assert_parse("a", symbol("a"))
  end

  def test_02_symbols
    assert_parse("a", symbols("a"))

    assert_parse("0", symbols("0".."9"))
    assert_parse("4", symbols("0".."9"))
    assert_parse("9", symbols("0".."9"))

    assert_parse("1", symbols("0".."9", "a".."f"))
    assert_parse("8", symbols("0".."9", "a".."f"))
    assert_parse("a", symbols("0".."9", "a".."f"))
    assert_parse("f", symbols("0".."9", "a".."f"))
    
    assert_parse("a", symbols("a", "c".."d", "k"))
    assert_parse("c", symbols("a", "c".."d", "k"))
    assert_parse("d", symbols("a", "c".."d", "k"))
    assert_parse("k", symbols("a", "c".."d", "k"))
  end

  def test_03_seq
    assert_parse("ab", seq(symbol("a"), symbol("b")))

    assert_parse("ab", seq("a", "b"))

    assert_parse("ak", seq("a".."d", "k"))
    assert_parse("bk", seq("a".."d", "k"))
    assert_parse("ck", seq("a".."d", "k"))
    assert_parse("dk", seq("a".."d", "k"))
  end

  def test_04_string
    assert_parse("def", string("def"))

    assert_parse("'\"'", string("'\"'"))
  end

  def test_05_seq_with_string_args
    assert_parse("code a", seq("code", " ", "a".."z"))
    assert_parse("code g", seq("code", " ", "a".."z"))
    assert_parse("code p", seq("code", " ", "a".."z"))
    assert_parse("code z", seq("code", " ", "a".."z"))
  end
  
  @@digit = symbols("0".."9")

  def test_06_choice
    assert_parse("1", choice(@@digit, seq("a", @@digit)))
    assert_parse("a9", choice(@@digit, seq("a", @@digit)))

    assert_parse("3.4", choice(seq(@@digit, ".", @@digit), @@digit)) 

    assert_parse("**", choice(string("**="), string("**")))
  end

  def test_07_repeat
    assert_parse("1", repeat(@@digit))
    assert_parse("12345", repeat(@@digit))
    p1 = repeat(@@digit)
    assert_parse("", repeat(@@digit))
  end

  def test_08_case_insesitive_parsing
    p = symbol("a")
    assert_exception(NoMatchError) {p.parse "A"}
    p.case_sensitive = false
    assert_parse("a", p)
    assert_parse("A", p)
    p.case_sensitive = true
    assert_parse("a", p)
    assert_exception(NoMatchError) {p.parse "A"}

    p2 = string "abc"
    assert_parse("abc", p2)
    assert_exception(NoMatchError) {p2.parse "ABC"}
    p2.case_sensitive = false
    assert_parse("ABC", p2)

    p3 = choice(symbol("a"), symbol("b"))
    p3.case_sensitive = false
    assert_parse("a", p3)
    assert_parse("A", p3)
    assert_parse("b", p3)
    assert_parse("B", p3)    
  end

  def test_09_apply_parser
    digit = symbols("0".."9")
    p = plus(digit)
    assert_equals(false, p.result_producing)

    p2 = p.apply(ToIntFunction.new)
    assert_kind_of(ApplyParser, p2)
    assert_equals(true, p2.result_producing)

    p3 = p.apply(proc{|x| x})
    assert_kind_of(ApplyParser, p3)
    assert_equals(true, p3.result_producing)
  end

  def test_10_premature_eof
    p = string("**=")
    assert_exception(UnexpectedEosError) {
      p.parse "**"
    }
  end

  def test_11_not_followed_by
    digit = symbols("0".."9")
    p = string(".").not_followed_by(digit)
    assert_parse(".a", p, ".")
    assert_exception(UnexpectedError) {p.parse ".1"}

    p2 = string(".").not_followed_by(digit, "no floating-point numbers")
    begin
      p2.parse ".1"
    rescue CustomParseError => e
      e.message =~ /(no floating-point numbers)/
      assert_equals("no floating-point numbers", $1)
    end

    p3 = string("-").not_followed_by(digit)
    assert_parse("-", p3, "-")
  end

  def test_12_skipparser_initialize
    p = SkipParser.new(symbol("a"))
    assert_kind_of(Parser, p)
    assert_kind_of(SkipParser, p)
  end

  def test_13_skipparser_match
    p = skip("a")
    assert_exception(NoMatchError) {p.parse("b")}
    assert_equals("", p.parse("a"))
    assert_equals("", p.parse("a!"))

    p2 = skip("0", choice("x", "X"))
    assert_exception(NoMatchError) {p2.parse("not")}
    assert_equals("", p2.parse("0x"))
    assert_equals("", p2.parse("0X"))
  end

  def test_14_fail
    p = failp("this parser always fails")
    assert_exception(CustomParseError) {
      p.parse "a"
    }
  end

  def test_15_raises
    p = symbol("a").raises("You've sent me 'a'!")
    assert_exception(CustomParseError) {
      p.parse "a"
    }    
  end

  def test_16_or_raises
    # or_raise and raise can be used for custom error messages, ie.
    # they will not be of the standard for "unexpected x when expecting y"
    # This can be useful to give error messages that are more to the point.

    msg = "Hexadecimal literal without digits"
    hexdigits = plus(symbols("0".."9", "a".."f"))
    hexlit = seq(skip("0x"), hexdigits)
    p = hexlit.or_raises(msg)

    assert_equals("34", hexdigits.parse("34"))
    assert_equals("34ef", hexlit.parse("0x34ef"))

    assert_equals("34ef", p.parse("0x34ef"))

    assert_exception(CustomParseError) {
      p.parse "0xq"
    }
    assert_exception(CustomParseError) {
      p.parse "q"
    }
  end

  def test_17_plus_around_maybe
    assert_parse("a", plus(maybe("a")))
    assert_parse("b", plus(maybe("a")), "")
  end

  def test_18_lookback_with_string
    p = seq(look_back(" \t"), "a") 
    st = OptimizedParseState.new " \ta"
    st.position = 2
    assert_equals("a", p.parse(st))

    st2 = OptimizedParseState.new "..a"
    st2.position = 2
    assert_exception(NoMatchError) {p.parse(st2)}

    p2 = seq(look_back("\n", true), "a")
    assert_equals("a", p2.parse("a"))
  end

  def test_19_lookback_with_symbols_parser
    white_space = symbols(" ", "\t")
    p = seq(look_back(white_space), "a") 
    st = OptimizedParseState.new "\ta"
    st.position = 1
    assert_equals("a", p.parse(st))
  end

  def test_20_none_of
    p = plus(none_of("\n"))
    assert_equals("# A small comment", p.parse("# A small comment\n"))

    p2 = plus(none_of("t\n"))
    assert_equals("# A small commen", p2.parse("# A small comment\n"))
  end

  def test_21_one_of
    p = plus(one_of("abc"))
    assert_equals("abcaabbcc", p.parse("abcaabbccd"))
  end

  def test_22_newline
    p = newline("a")
    assert_kind_of(Parser, p)
    assert_kind_of(NewlineParser, p)
  end

  def test_23_any_until
    # Default is to match and return the end marker
    p = any_until("a")
    assert_equals("bbbba", p.parse("bbbba")) 

    # If we don't want to return it we specify with a skip. Note that the a
    # is still consumed.
    p = any_until(skip("a"))
    assert_equals("bbbb", p.parse("bbbba")) 

    # If we don't want to consume the end marker we give false as 2nd param.
    p = any_until("a", false)
    assert_equals("bbbb", p.parse("bbbba"))

    # Should fail unless it actually finds the end
    p = any_until("a", false)
    assert_exception(UnexpectedEosError) {p.parse "b"}
  end

  def test_24_look_ahead
    p = seq("a", look_ahead(" "))

    st = OptimizedParseState.new "a "
    st.position = 0
    assert_equals("a", p.parse(st))
    assert_equals(1, st.position)

    #assert_exception(ParseError) {p.parse "a"}
  end

  def test_25_lift_result
    p = seq("a", plus(symbols("0".."9")).apply(ToIntFunction.new)).lift(1)
    assert_equals(1, p.parse("a1"))
  end

  def test_26_capture_and_ref
    p = seq(capture(one_of("ba"), :f), plus(symbols("0".."9")), ref(:f))
    assert_equals("a123a", p.parse("a123a"))
    assert_equals("b1b",   p.parse("b1b"))
  end

  def test_27_succeed
    p = succeed("")
    assert_equals("", p.parse("ara"))

    p = succeed
    assert_equals("", p.parse("ara"))
  end

  def test_28_skipping_when_partial_match
    p = choice(seq(skip("<<"), "a"),
	       seq("<<", "b"))
    assert_equals("<<b", p.parse("<<b"))
  end

  def test_29_recursive
    rp = recursive("(", choice(".", recurse), ")")
    assert_kind_of(Parser, rp)
    assert_kind_of(RecursiveParser, rp)

    assert_equals("(.)", rp.parse("(.)"))
    assert_equals("(.)", rp.parse("(.)("))
    assert_equals("(.)", rp.parse("(.)a"))
    assert_equals("((.))", rp.parse("((.))("))
    
    assert_exception(NoMatchError) {rp.parse("c")}
    assert_exception(NoMatchError) {rp.parse(".")}
    assert_exception(UnexpectedEosError) {rp.parse("(")}
    assert_exception(NoMatchError) {rp.parse(")")}
    assert_exception(NoMatchError) {rp.parse("()")}
  end

  def delimited(open, close, nested = true)
    if nested
      not_open_or_close = none_of(open + close)
      recp = recursive(open, mult(mult(not_open_or_close), recurse), 
		       close)
      inner = mult(mult(not_open_or_close), recp)
    else
      inner = mult(none_of(close))
    end
    
    seq(skip(open), inner, skip(close))
  end

  def test_30_recursive_and_nested
    dp = seq(skip("%r"), choice(delimited("{", "}"),
				delimited("<", ">")))

    assert_equals("abcd", dp.parse("%r{abcd}"))
    assert_equals("ab{cd}{e}{{f}}", dp.parse("%r{ab{cd}{e}{{f}}}"))
    assert_equals("abcd", dp.parse("%r<abcd>"))
    assert_equals("ab<cd><e><<f>>", dp.parse("%r<ab<cd><e><<f>>>"))
  end

  def test_31_notp
    marker = choice("/*", "*/")
    p = notp(marker)
    assert_equals("a", p.parse("a"))
    p = notp(marker, false)
    assert_equals("", p.parse("a"))
    assert_exception(UnexpectedError) {p.parse("/*")}
  end

  def test_32_separatingsequenceparser
    p = sepseq("a", "1", "*")
    assert_equals(["a", "1", "*"], p.parse("a1*"))
  end

  def test_33_rdg_basic
    print_command = string "PRINT"
    assert_equals("PRINT", print_command.parse("PRINT"))

    valid_ascii = (0x20..0x7e).to_a.select {|c| c != '"'[0] and c != "\\"[0]}
    string_character = choice(*valid_ascii.map{|c| string(c.chr)})
    dquote = string '"'
    string_literal = seq(dquote, mult(string_character), dquote)
    assert_equals('"Hello, world!"', 
		  string_literal.parse('"Hello, world!"'))

    white_space = skip(mult(choice(string(" "), string("\t"))))
    command = seq(print_command, white_space, string_literal)
    # Difference to the book is that seq's by the default returns their
    # results as one large string. If you want them separated you must
    # specify it explicitly.
    assert_equals('PRINT"a"', command.parse('PRINT "a"'))

    p = sepseq(print_command, white_space, string_literal)
    assert_equals(['PRINT', '"a"'], p.parse('PRINT "a"'))
  end

  def test_34_nested_parenthesis
    parens = recursive("(" & maybe(recurse) & ")")
    assert_equals("((()))", parens.parse("((()))"))
    assert_exception(NoMatchError) {
      parens.parse("((()a")
    }
  end

  def test_35_seq_into_sepseq
    p = seq("a", "1")
    p2 = separated_results(p)
    assert_equals(["a", "1"], p2.parse("a1"))
  end

  def test_36_building_term
    atree = Term.new_sub_class(:Assignment, :variable, :value)
    etree = Term.new_sub_class(:Expression, :left,     :right)

    ident = plus(choice("a".."z", "A".."Z"))
    integer = plus("0".."9")
    expression = (integer & skip("+") & integer).build(etree)
    assignment = (ident & skip("=") & expression).build(atree)
    
    res = assignment.parse "temp=1+2"
    assert_kind_of(Term, res)
    assert_kind_of(atree, res)

    assert(res.respond_to?(:variable))
    assert_equals("temp", res.variable)

    assert(res.respond_to?(:value))
    valuetree = res.value

    assert_kind_of(Term,  valuetree)
    assert_kind_of(etree, valuetree)

    assert(valuetree.respond_to?(:left))
    assert_equals("1", valuetree.left)

    assert(valuetree.respond_to?(:right))
    assert_equals("2", valuetree.right)    
  end

  def test_37_plus_of_resultproducing_parsers
    it = Term.new_sub_class(:It, :v)
    p = plus(inner = string("a").build(it))
    assert_equals(true, inner.result_producing)
    assert_equals(true, p.result_producing)
    num = 3
    res = p.parse("a" * num)
    assert_kind_of(Array, res)
    assert_equals(num, res.length)
    num.times {|i| assert_kind_of(it, res[i])}
  end

  def test_38_result_producing
    it = Term.new_sub_class(:It, :v)    
    assert_equals(true, choice("a", string("b").build(it)).result_producing)
    # Should test that it forces a result from the "a" parser even though
    # it will not produce a result by itself...
  end

  def test_39_position_resetted_after_repeat
    int = plus(Digit)
    exp = plus(Letter)
    statement = int & "=" & exp
    label = mult(int & ":") 
    labeled_statement = label & statement
    st = OptimizedParseState.new "1:2=a"
    label.parse(st)
    assert_equals(2, st.position)

    # Fixing the bug above gave new bug when min and max were both specified:
    p = repeat(Digit, 1, 2)
    assert_equals("01", p.parse("01"))
  end

  def test_40_repeated_recurse_in_temp_var
    # This works:
    p = recursive("a" & (me = recurse) & "b" & me | "1")
    assert_equals("a1b1", p.parse("a1b1"))

    # This didn't but now does:
    p = recursive("a" & (me = plus(recurse)) & "b" & me | "1")
    assert_equals("a1b1", p.parse("a1b1"))
    assert_equals("a11b111", p.parse("a11b111"))
  end

  def test_41_one_of_with_string_of_length_1
    assert_kind_of(OneOfParser, one_of("a"))
  end

  def assert_first_set(strOrCharSet, parser)
    if String === strOrCharSet
      cs = CharSet.new(strOrCharSet)
    else
      cs = strOrCharSet
    end
    assert_equals(cs, fs = parser.first_set)
  end

  def test_42_first_set
    assert_first_set("a",    one_of("a"))
    assert_first_set("abc",  one_of("cab"))
    assert_first_set("ab",   choice("a", "b"))
    assert_first_set("a",    seq("a", "b"))
    assert_first_set("ab",   seq(choice("a", "b"), "c"))
    assert_first_set("ab",   seq(maybe("a"), "b"))
    assert_first_set(CharSet.new("ab", true),      seq(maybe("a"), maybe("b")))
    assert_first_set("ab",   seq(mult("a"), "b"))
    assert_first_set(CharSet.new("ab", true),      seq(mult("a"), mult("b")))
    assert_first_set("a",    seq(plus("a"), "b"))
    assert_first_set("a",    seq(plus("a"), plus("b")))
    assert_first_set(CharSet.new("a", false).invert, notp("a"))
    assert_first_set("a",    skip("a"))
    assert_first_set(CharSet.new("a").invert, none_of("a"))
    assert_first_set(CharSet.new("", true), look_back("a"))
    assert_first_set(CharSet.new("", true), look_ahead("a"))
    assert_first_set(CharSet.new("a").invert, any_until("a"))
  end
end
