require 'rubyunit'

require 'rockit/parse/parse_error'
include Rockit::Parse

MockParser = Struct.new("Parser", :expected)
MockState = Struct.new("ParserState", :position, :current)

class ParseErrorTest < TestCase
  def assert_parse_error(unexp, exp, pos, unexpmsg, expmsg = nil)
    pe = ParseError.new(unexp, exp, pos)
    assert_kind_of(ParseError, pe)
    assert_message(pe, unexpmsg, expmsg)
  end

  def assert_message(error, unexpmsg, expmsg)
    msg = error.message
    assert_kind_of(String, msg)
    msg =~ eval("/^unexpected (" + Regexp.escape(unexpmsg) + ")/")
    assert_equals(unexpmsg.length, $1.length)
    if expmsg
      msg =~ eval("/^when expecting (" + Regexp.escape(expmsg) + ")/")
      assert_equals(expmsg.length, $1.length)
    end
  end
  
  @@eos = "end of stream"

  def test_01_parse_error
    assert_parse_error(@@eos, [],           0,   @@eos)
    assert_parse_error(@@eos, [?a],         0,   @@eos, "'a'")
    assert_parse_error(@@eos, [?1, ?*],     0,   @@eos, "'1' or '*'")
    assert_parse_error(?a,    [?_],         100, "'a'", "'_'")
    assert_parse_error(?_,    [?z, ?9, ?=], 34,  "'_'", "'z', '9' or '='")
  end

  def assert_unexpeos_error(expected, pos, expmsg)
    p = MockParser.new(expected)
    s = MockState.new(pos)
    ue = UnexpectedEosError.new(p,s)
    assert_kind_of(UnexpectedEosError, ue)
    assert_kind_of(ParseError, ue)
    assert_message(ue, @@eos, expmsg)
  end

  def test_02_unexpected_eos_error
    assert_unexpeos_error([?a], 0, "'a'")
    assert_unexpeos_error([?a, ?b], 0, "'a' or 'b'")
    assert_unexpeos_error([?a, ?b, ?_], 0, "'a', 'b' or '_'")
  end

  def assert_nomatch_error(expected, pos, current, expmsg)
    p = MockParser.new(expected)
    s = MockState.new(pos, current)
    nme = NoMatchError.new(p,s)
    assert_kind_of(NoMatchError, nme)
    assert_kind_of(ParseError, nme)
    assert_message(nme, "'#{current.chr}'", expmsg)
  end

  def test_03_nomatch_error
    assert_nomatch_error([?a], 0, ?b, "'a'")
    assert_nomatch_error([?a, ?c], 0, ?b, "'a' or 'c'")
    assert_nomatch_error([?1, ?_, ?w], 0, ?*, "'1', '_' or 'w'")
  end

  def test_04_parseerror_merge
    p1 = ParseError.new(?a, [?b], 0)
    p2 = ParseError.new(?a, [?c], 0)
    p = p1.merge(p2)
    assert_kind_of(ParseError, p)
    assert_message(p, "'a'", "'b' or 'c'")

    p1 = ParseError.new(?a, [?b], 0)
    p2 = ParseError.new(?a, [], 0)
    p = p1.merge(p2)
    assert_kind_of(ParseError, p)
    assert_message(p, "'a'", "'b'")

    p1 = ParseError.new(?d, [?c], 2)
    p2 = ParseError.new(?b, [?c], 1)
    p = p1.merge(p2)
    assert_kind_of(ParseError, p)
    assert_message(p, "'d'", "'c'")

    p1 = ParseError.new(?b, [?c], 1)
    p2 = ParseError.new(?d, [?c], 2)
    p = p1.merge(p2)
    assert_kind_of(ParseError, p)
    assert_message(p, "'d'", "'c'")
  end

  def test_05_expected_is_strings
    assert_parse_error(@@eos, ["a letter"],         0,   @@eos, "a letter")
    assert_parse_error(@@eos, ["a letter", ?1],     0,   @@eos, "a letter or '1'")
    assert_parse_error(@@eos, [?1, "a letter", ?_], 0,   @@eos, "'1', a letter or '_'")
  end

  def assert_unexp_error(unexp, pos, unexpmsg)
    s = MockState.new(pos, unexp)
    ue = UnexpectedError.new(s)
    assert_kind_of(UnexpectedError, ue)
    assert_kind_of(ParseError, ue)
    assert_message(ue, unexpmsg, nil)
  end

  def test_06_unexpected_error
    assert_unexp_error(?a, 0, "'a'")
    assert_unexp_error(?1, 0, "'1'")
    assert_unexp_error(?*, 0, "'*'")
  end

  def test_07_custom_parse_error
    s = MockState.new(0)    
    cpe = CustomParseError.new("my message", s)
    assert_kind_of(CustomParseError, cpe)
    cpe.message =~ /(my message)/
    assert_equals("my message", $1)
  end

  def make_ueos(expected, pos)
    p = MockParser.new(expected)
    s = MockState.new(pos)
    UnexpectedEosError.new(p,s)
  end

  def test_08_merging_unexpectedeos
    ue1, ue2 = make_ueos(["1"], 0), make_ueos(["2"], 0)
    u = ue1.merge(ue2)
    assert_kind_of(UnexpectedEosError, u)
  end
end

