require 'rubyunit'

require 'ruby_lexer'
include Ruby

module Ruby
  class ModedState
    def position=(newPos); @stream.position = newPos; end
  end
end

class TestRubyLexer < TestCase
	
  #---------------------- Individual Token Tests ---------------------------

  def test_01_initialize
    t = Lexer.new ""
    assert_kind_of(Lexer, t)
  end

  def test_02_end_of_stream
    assert_tok("", false, false)
  end
	
  def test_03_whitespace
    assert_tok(" 1",         :tINTEGER, 1)
    assert_tok("\t 1",       :tINTEGER, 1)
    assert_tok("\f \t1",     :tINTEGER, 1)
    assert_tok("\r \t\f1",   :tINTEGER, 1)
    assert_tok("\v \t\f\r1", :tINTEGER, 1)
    assert_tok(" \13a",      :tIDENTIFIER, "a") # RACK
  end

  def test_04_line_comment 
    # This is a deviation from Matz yylex which simply skips comments.
    # We can't do that since people will want to be notified of comments
    # to extract info from them, colour them etc.
    assert_tok("# A comment\n", :tComment, "# A comment", :postst => EXPR_BEG)

    assert_tok("# A comment", :tComment, "# A comment", :postst => EXPR_BEG)
  end

  def test_05_block_comment # =begin...=end
    rd_doc = <<-EOS
=begin
     Some crazy stuf...!
 test
=end
    EOS
    assert_tok(rd_doc, :tBlockComment, "     Some crazy stuf...!\n test\n")
  end

  def test_06_newline
    assert_tok("\n1", :tINTEGER, 1, 
	       :prest => [EXPR_BEG, EXPR_FNAME, EXPR_DOT]) 

    assert_tok("\na", "\n", "\n", 
	       :prest => [EXPR_END, EXPR_ARG, EXPR_CLASS, EXPR_MID],
	       :postst => EXPR_BEG) 
  end

  def test_07_star # *
    assert_tok("**=", :tOP_ASGN, "**=", :postst => EXPR_BEG, :info => :tPOW)
    assert_tok("**", :tPOW, "**", :postst => EXPR_BEG)
    assert_tok("*=", :tOP_ASGN, "*=", :postst => EXPR_BEG, :info => "*")
    assert_tok("*", :tSTAR, "*", :prest => EXPR_BEG, :postst => EXPR_BEG)
    assert_tok("*", :tSTAR, "*", :prest => EXPR_MID, :postst => EXPR_BEG)
    assert_tok("*", "*", "*", :prest => (EXPR_STATES - [EXPR_BEG, EXPR_MID]), 
	       :postst => EXPR_BEG)
  end

  def test_08_exclamation # !
    assert_tok("!=", :tNEQ, "!=", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("!~", :tNMATCH, "!~", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    assert_tok("!", "!", "!", :prest => EXPR_STATES, :postst => EXPR_BEG)
  end

  def test_09_equal_sign # =
    assert_tok("===", :tEQQ, "===", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("==",  :tEQ,  "==", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("=~",  :tMATCH,  "=~", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    assert_tok("=>",  :tASSOC,  "=>", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    assert_tok("=",  "=",  "=", :prest => EXPR_STATES, :postst => EXPR_BEG)    
  end

  def test_10_less_than
    assert_tok("<=>", :tCMP, "<=>", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("<=", :tLEQ, "<=", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("<<=", :tOP_ASGN, "<<=", :info => :tLSHFT,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("<<", :tLSHFT, "<<", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    assert_tok("<", "<", "<", :prest => EXPR_STATES, :postst => EXPR_BEG)
  end

  def test_11_here_document
    assert_tok("<<EOS\na\nEOS", :tDSTRING, "a\n", 
    	       :prest => EXPR_STATES - [EXPR_END, EXPR_CLASS, EXPR_ARG])
    assert_tok("<<'EOS'\na\nEOS", :tSTRING, "a\n", 
    	       :prest => EXPR_STATES - [EXPR_END, EXPR_CLASS, EXPR_ARG])
    assert_tok("<<`EOS`\na\nEOS", :tSTRING, "a\n", 
    	       :prest => EXPR_STATES - [EXPR_END, EXPR_CLASS, EXPR_ARG])

    assert_tok("<<-EOS\na\n  EOS", :tDSTRING, "a\n", 
    	       :prest => EXPR_STATES - [EXPR_END, EXPR_CLASS, EXPR_ARG])

    assert_tok(" <<EOS\na\nEOS", :tDSTRING, "a\n", :prest => EXPR_ARG)
  end

  def test_12_greater_than
    assert_tok(">=", :tGEQ, ">=", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok(">>=", :tOP_ASGN, ">>=", :info => :tRSHFT,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok(">>", :tRSHFT, ">>", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok(">", ">", ">", :prest => EXPR_STATES, :postst => EXPR_BEG)
  end

  def test_13_question_mark
    assert_tok("? ", "?", "?", :prest => END_AND_ENDARG, :postst => EXPR_BEG)

    #assert_exception(RuntimeError) {assert_tok("?", :dummy, :dummy)}
    #assert_exception(RuntimeError) {
    #  assert_tok("?", :dummy, :dummy, :prest => EXPR_ARG)
    #}

    assert_tok("?a", :tINTEGER, ?a, :prest => EXPR_STATES - END_AND_ENDARG,
	       :postst => EXPR_END)
    assert_tok("??", :tINTEGER, ??, :prest => EXPR_STATES - END_AND_ENDARG,
	       :postst => EXPR_END)

    assert_tok("?\\xa", :tINTEGER, ?\xa, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)
    assert_tok("?\\x01", :tINTEGER, ?\x01, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)

    assert_tok("?\\1", :tINTEGER, ?\1, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)
    assert_tok("?\\72", :tINTEGER, ?\72, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)
    assert_tok("?\\345", :tINTEGER, ?\345, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)

    "ntrfvaebs".each_byte do |b|
      s = "?\\X"; s[2] = b      
      assert_tok(s, :tINTEGER, eval(s),
		 :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)
    end
  end

  def test_14_ampersand
    assert_tok("&&=", :tOP_ASGN, "&&=", :info => :tANDOP,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("&&", :tANDOP, "&&", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("&=", :tOP_ASGN, "&=", :info => "&",
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok(" &a", :tAMPER, "&", :prest => EXPR_ARG, :postst => EXPR_BEG) 
    assert_tok("&a", "&", "&", :prest => EXPR_ARG, :postst => EXPR_BEG) 
    assert_tok("&", :tAMPER, "&", :prest => [EXPR_BEG, EXPR_MID],
	       :postst => EXPR_BEG) 
    assert_tok("&", "&", "&", 
	       :prest => EXPR_STATES - [EXPR_BEG, EXPR_MID, EXPR_ARG],
	       :postst => EXPR_BEG) 
  end

  def test_15_bar
    assert_tok("||=", :tOP_ASGN, "||=", :info => :tOROP,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("||", :tOROP, "||", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("|=", :tOP_ASGN, "|=", :info => "|",
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("|", "|", "|", :prest => EXPR_STATES, :postst => EXPR_BEG) 
  end

  def test_16_plus
    assert_tok("+@", :tUPLUS, "+@", :prest => [EXPR_FNAME, EXPR_DOT])
    assert_tok("+", "+", "+", :prest => [EXPR_FNAME, EXPR_DOT])
    assert_tok("+=", :tOP_ASGN, "+=", :info => "+",
	       :prest => EXPR_STATES - [EXPR_FNAME, EXPR_DOT], 
	       :postst => EXPR_BEG)
    assert_tok("+9876", :tINTEGER, 9876, :postst => EXPR_BEG) #RACK
    assert_tok("+", "+", "+", :prest => NOT_FNAME_OR_DOT,
	       :postst => EXPR_BEG) #RACK
    # arg_ambigous:
    #assert_tok(" +1", :tINTEGER, 1, :prest => EXPR_ARG,
    #	       :postst => EXPR_BEG)
		
  end

  def test_17_minus
    assert_tok("-@", :tUMINUS, "-@", :prest => [EXPR_FNAME, EXPR_DOT])
    assert_tok("-", "-", "-", :prest => [EXPR_FNAME, EXPR_DOT])
    assert_tok("-=", :tOP_ASGN, "-=", :info => "-",
	       :prest => EXPR_STATES - [EXPR_FNAME, EXPR_DOT], 
	       :postst => EXPR_BEG)
    assert_tok("-12345", :tINTEGER, -12345)
    # More tests needed...
  end

  def test_18_dot
    assert_tok("...", :tDOT3, "...", :prest => EXPR_STATES, :postst => EXPR_BEG) #RACK
    assert_tok("..", :tDOT2, "..", :prest => EXPR_STATES, :postst => EXPR_BEG) #RACK
    assert_tok(".a", ".", ".", :prest => EXPR_STATES, :postst => EXPR_DOT)
    assert_tok(".5", :tFLOAT, 0.5)
  end

  def test_19_number_integer #RACK
    assert_tok("2", :tINTEGER, 2)
    assert_tok("12", :tINTEGER, 12)
    assert_tok("-15", :tINTEGER, -15)
    assert_tok("12_000", :tINTEGER, 12000)
    assert_tok("987_6_5_432_10", :tINTEGER, 987_6_5_432_10)
  end
	
  def test_20_number_hex #RACK
    assert_tok("0xf", :tINTEGER, 15)
    assert_tok("0xab", :tINTEGER, 10*16+11)
    assert_tok("0x__c_d_e", :tINTEGER, 12*16*16+13*16+14)
    assert_tok("0xf", :tINTEGER, 0xf)
    assert_tok("0xab", :tINTEGER, 0xa_b)
    assert_tok("0x__c_d_e", :tINTEGER, 0x__c_d_e)
  end
	
  def test_21_number_binary #RACK
    assert_tok("0b01", :tINTEGER, 0b01)
    assert_tok("0b_11_0", :tINTEGER, 0b_11_0)
  end
	
  def test_22_number_octal #RACK
    assert_tok("0123", :tINTEGER, 0123)
    assert_tok("0_7_6_5", :tINTEGER, 0_7_6_5)
  end

  def test_23_number_float #RACK
    assert_tok("1.2345", :tFLOAT, 1.2345)
    assert_tok("-1e10", :tFLOAT, -1e10)
    assert_tok("-9.87E-1", :tFLOAT, -9.87E-1)
    assert_tok("-.2345", :tFLOAT, -.2345)
  end

  def test_24_colon
    assert_tok(":test", :tSYMBEG, ":")
    # TODO: Add more colon tests...
  end

  def test_25_identifier
    assert_tok("my_ident", :tIDENTIFIER, "my_ident")
  end

  def test_26_constant
    assert_tok("My_ident_", :tCONSTANT, "My_ident_")    
  end

  def test_27_fid
    assert_tok("a?",  :tFID, "a?")
    assert_tok("z!",  :tFID, "z!")
  end

  def test_28_instance_variable
    assert_tok("@a",  :tIVAR, "@a")
    assert_tok("@z_", :tIVAR, "@z_")
  end

  def test_29_class_variable
    assert_tok("@@a",  :tCVAR, "@@a")
    assert_tok("@@z_", :tCVAR, "@@z_")
  end

  def test_30_global_variable
    assert_tok("$a",   :tGVAR, "$a", :postst => EXPR_END)
    assert_tok("$z_",  :tGVAR, "$z_", :postst => EXPR_END)
    assert_tok("$_za", :tGVAR, "$_za", :postst => EXPR_END)

    assert_tok("$*",  :tGVAR, "$*", :postst => EXPR_END)
    assert_tok("$$",  :tGVAR, "$$", :postst => EXPR_END)
    assert_tok("$?",  :tGVAR, "$?", :postst => EXPR_END)
    assert_tok("$!",  :tGVAR, "$!", :postst => EXPR_END)
    assert_tok("$@",  :tGVAR, "$@", :postst => EXPR_END)
    assert_tok("$/",  :tGVAR, "$/", :postst => EXPR_END)
    assert_tok("$\\", :tGVAR, "$\\", :postst => EXPR_END)
    assert_tok("$;",  :tGVAR, "$;", :postst => EXPR_END)
    assert_tok("$,",  :tGVAR, "$,", :postst => EXPR_END)
    assert_tok("$.",  :tGVAR, "$.", :postst => EXPR_END)
    assert_tok("$=",  :tGVAR, "$=", :postst => EXPR_END)
    assert_tok("$:",  :tGVAR, "$:", :postst => EXPR_END)
    assert_tok("$<",  :tGVAR, "$<", :postst => EXPR_END)
    assert_tok("$>",  :tGVAR, "$>", :postst => EXPR_END)
    assert_tok("$\"", :tGVAR, "$\"", :postst => EXPR_END)
  end

  def test_31_back_ref
    assert_tok("$&",   :tBACK_REF, "&", :postst => EXPR_END)
    assert_tok("$`",   :tBACK_REF, "`", :postst => EXPR_END)
    assert_tok("$'",   :tBACK_REF, "'", :postst => EXPR_END)
    assert_tok("$+",   :tBACK_REF, "+", :postst => EXPR_END)
  end

  def test_32_nth_ref
    assert_tok("$1",    :tNTH_REF, 1, :postst => EXPR_END)  #RACK
    assert_tok("$9999", :tNTH_REF, 9999, :postst => EXPR_END) #RACK
  end

  def test_33_reserved_word
    # Positive tests
    assert_tok("end",      :kEND,        "end",      :postst => EXPR_END)    
    assert_tok("else",     :kELSE,       "else",     :postst => EXPR_BEG)    
    assert_tok("case",     :kCASE,       "case",     :postst => EXPR_BEG)    
    assert_tok("ensure",   :kENSURE,     "ensure",   :postst => EXPR_BEG)    
    assert_tok("module",   :kMODULE,     "module",   :postst => EXPR_BEG)    
    assert_tok("elsif",    :kELSIF,      "elsif",    :postst => EXPR_BEG)    
    assert_tok("def",      :kDEF,        "def",      :postst => EXPR_FNAME)    
    assert_tok("rescue",   :kRESCUE,     "rescue",   
	       :prest => EXPR_BEG, :postst => EXPR_MID)    
    assert_tok("rescue",   :kRESCUE_MOD, "rescue",   
	       :prest => EXPR_STATES - [EXPR_BEG], :postst => EXPR_BEG)    
    assert_tok("not",      :kNOT,        "not",      :postst => EXPR_BEG)    
    assert_tok("then",     :kTHEN,       "then",     :postst => EXPR_BEG)    
    assert_tok("yield",    :kYIELD,      "yield",    :postst => EXPR_ARG)    
    assert_tok("for",      :kFOR,        "for",      :postst => EXPR_BEG)    
    assert_tok("self",     :kSELF,       "self",     :postst => EXPR_END)    
    assert_tok("false",    :kFALSE,      "false",    :postst => EXPR_END)    
    assert_tok("retry",    :kRETRY,      "retry",    :postst => EXPR_END)    
    assert_tok("return",   :kRETURN,     "return",   :postst => EXPR_MID)    
    assert_tok("true",     :kTRUE,       "true",     :postst => EXPR_END)    
    assert_tok("if",       :kIF,         "if",       :postst => EXPR_BEG)    
    assert_tok("if",       :kIF_MOD,     "if",   :prest => EXPR_END, :postst => EXPR_BEG)    
    #assert_tok("defined?", :kDEFINED,    "defined?", :postst => EXPR_ARG)    
    assert_tok("super",    :kSUPER,      "super",    :postst => EXPR_ARG)    
    assert_tok("undef",    :kUNDEF,      "undef",    :postst => EXPR_FNAME)    
    assert_tok("break",    :kBREAK,      "break",    :postst => EXPR_END)    
    assert_tok("in",       :kIN,         "in",       :postst => EXPR_BEG)    
    assert_tok("do",       :kDO,         "do",       :postst => EXPR_BEG)    
    #assert_tok("do",       :kDO_COND,    "do",  :postst => EXPR_BEG)  {|t| t.cond_push}
    #assert_tok("do",       :kDO_BLOCK,   "do",  :postst => EXPR_BEG)  {|t| t.cmdarg_push}
    assert_tok("nil",      :kNIL,        "nil",      :postst => EXPR_END)    
    assert_tok("until",    :kUNTIL,      "until",    :postst => EXPR_BEG)
    assert_tok("until",    :kUNTIL_MOD,  "until",    :prest => EXPR_END, :postst => EXPR_BEG)    
    assert_tok("unless",   :kUNLESS,     "unless",   :postst => EXPR_BEG)    
    assert_tok("unless",   :kUNLESS_MOD, "unless",  :prest => EXPR_END,   :postst => EXPR_BEG)    
    assert_tok("or",       :kOR,         "or",       :postst => EXPR_BEG)    
    assert_tok("next",     :kNEXT,       "next",     :postst => EXPR_END)    
    assert_tok("when",     :kWHEN,       "when",     :postst => EXPR_BEG)    
    assert_tok("redo",     :kREDO,       "redo",     :postst => EXPR_END)    
    assert_tok("and",      :kAND,        "and",      :postst => EXPR_BEG)    
    assert_tok("begin",    :kBEGIN,      "begin",    :postst => EXPR_BEG)    
    assert_tok("__LINE__", :k__LINE__,   "__LINE__", :postst => EXPR_END)    
    assert_tok("__FILE__", :k__FILE__,   "__FILE__", :postst => EXPR_END)    
    assert_tok("class",    :kCLASS,      "class",    :postst => EXPR_CLASS)    
    assert_tok("END",      :klEND,       "END",      :postst => EXPR_END)    
    assert_tok("BEGIN",    :klBEGIN,     "BEGIN",    :postst => EXPR_END)    
    assert_tok("while",    :kWHILE,      "while",    :postst => EXPR_BEG)    
    assert_tok("alias",    :kALIAS,      "alias",    :postst => EXPR_FNAME)    

    # Negative tests
    assert_tok("enda",     :tIDENTIFIER,  "enda")
    assert_tok("alias1",   :tIDENTIFIER,  "alias1")
    assert_tok("and?",     :tFID,         "and?")
    assert_tok("BEGINA",   :tCONSTANT,    "BEGINA")
    assert_tok("__self__", :tIDENTIFIER,  "__self__")
  end
  
  def test_34_comma
    assert_tok(",", ",", ",", :prest => EXPR_STATES,
	       :postst => EXPR_BEG)
  end

  def test_35_semicolon
    assert_tok(";", ";", ";", :prest => EXPR_STATES,
	       :postst => EXPR_BEG)
  end

  def test_36_caret
    assert_tok("^=", :tOP_ASGN, "^=", :info => "^",
	       :prest => EXPR_STATES, :postst => EXPR_BEG)    
    assert_tok("^", "^", "^", :prest => EXPR_STATES,
	       :postst => EXPR_BEG)    
  end

  def test_37_tilde
    assert_tok("~@", "~", "~", #:postpos => 2, 
	       :prest => [EXPR_FNAME, EXPR_DOT], :postst => EXPR_BEG)    
    assert_tok("~", "~", "~", :postpos => 1,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)    
  end

  def test_38_slash
    assert_tok("/=", :tOP_ASGN, "/",
	       :prest => EXPR_STATES - [EXPR_BEG, EXPR_MID],
	       :postst => EXPR_BEG)
    # How does this interact with a normal Regexp?
    #assert_tok("/=/", :tREGEXP, "/=/", :prest => [EXPR_BEG, EXPR_MID])
    assert_tok("/", "/", "/", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    #assert_tok(" /a/", :tREGEXP, "/a/", :prest => EXPR_ARG)
    #assert_tok("/a/", "/", "/", :prest => EXPR_ARG, :postst => EXPR_BEG)
    #assert_tok(" / ", "/", "/", :prest => EXPR_ARG, :postst => EXPR_BEG)
  end

  def test_39_left_brace
    assert_tok("{",     :tLBRACE, "{", 
	       :prest => EXPR_STATES - [EXPR_END, EXPR_ARG],
	       :postst => EXPR_BEG)
    assert_tok("{",     "{", "{", 
	       :prest => [EXPR_END, EXPR_ARG], :postst => EXPR_BEG)
  end

  def test_40_right_brace
    assert_tok("}",  "}",  "}")
  end

  def test_41_left_bracket
    assert_tok("[", "[", "[", :prest => [EXPR_FNAME, EXPR_DOT]) #RACK
    assert_tok("[", :tLBRACK, "[", :prest => [EXPR_BEG, EXPR_MID],  :postst => EXPR_BEG) #RACK
    assert_tok("[", "[", "[", 
	       :prest => EXPR_STATES - FNAME_AND_DOT - BEG_AND_MID-[EXPR_ARG],
	       :postst => EXPR_BEG) #RACK

    # Special treatment since assert_tok can't set the position in the stream
    # Maybe extend it?
    st = Ruby::ModedState.new " ["
    st.mode = Ruby::EXPR_ARG
    st.position = 1
    l = Lexer.new st 
    tok = l.next_token
    assert_kind_of(RaccToken, tok)
    assert_equals(:tLBRACK, tok[0])
    assert_equals("[", tok[1])
  end

  def test_42_right_bracket
    assert_tok("]",  "]", "]")
  end

  def test_43_aref
    assert_tok("[]", :tAREF, "[]", :prest => EXPR_FNAME, :postst => EXPR_FNAME)
    assert_tok("[]",  :tAREF, "[]", :prest => EXPR_DOT, :postst => EXPR_DOT)
    assert_tok("[]=", :tLBRACK, "[", :prest => EXPR_BEG)
  end

  def test_44_aset
    assert_tok("[]=",   :tASET, "[]=", :prest => EXPR_FNAME, :postst => EXPR_FNAME)
    assert_tok("[]=",   :tASET, "[]=", :prest => EXPR_DOT, :postst => EXPR_DOT)
    assert_tok("[]=", :tLBRACK, "[", :prest => EXPR_BEG)
  end

  def test_45_left_parenthesis # RACK
    assert_tok("(", :tLPAREN, "(", :prest => [EXPR_BEG, EXPR_MID], 
	       :postst => EXPR_BEG)
    assert_tok("(", "(", "(", :prest => EXPR_STATES - [EXPR_BEG, EXPR_MID],
	       :postst => EXPR_BEG)
  end
	
  def test_46_right_parenthesis # RACK
    assert_tok(")",  ")",  ")", :postst => EXPR_END)
    #assert_tok("()",  ")",  ")", :postst => EXPR_END)
  end

  def test_47_doublequoted_string
    assert_tok("\"a\"",  :tDSTRING, "a")
    assert_tok("\"#a\"",  :tDSTRING, "#a")
    #assert_tok('%{"a"}', :tDSTRING, "\"a\"")
    #assert_tok('%|"a"|', :tDSTRING, "\"a\"")
  end

  def test_48_singlequoted_string
    assert_tok("'a'",  :tSTRING, "a")
    assert_tok("'#a'",  :tSTRING, "#a")
  end

  def test_49_regexp
    assert_tok("/a/",  :tREGEXP, "a")
    #assert_tok("/=/", :tREGEXP, "=") but depends on mode?? Check!
  end

  #---------------------- Simple Statement Tests ---------------------------
	
  def test_simple_statement_1
    assert_tok_list( "22 + 22", 
		     [ [:tINTEGER, 22], ["+", "+"], [:tINTEGER, 22] ] )
  end
	
  def test_simple_statement_2
    assert_tok_list( "x=23", 
		     [ [:tIDENTIFIER, "x"], ["=","="], [:tINTEGER, 23] ] )
  end

  #---------------------- Complex Statement Tests --------------------------

  def test_complex_statement_1
    assert_tok_list <<-EOS,
      def method1
	x=32
      end
    EOS
      [ 
	[:kDEF, "def"], [:tIDENTIFIER, "method1"],
	[:tIDENTIFIER, "x"], ["=","="], [:tINTEGER, 32],
	[:kEND, "end"], 
	["\n", "\n"] #RF added this one although unsure if its correct?
      ]
  end
  #-------------------- Assert Helper Functions -------------------------
	
  def assert_tok_list(string, tok_list)
    t = Lexer.new string
    t.state = EXPR_BEG
    tok_list.each do |token|
      parsed_token = t.next_token
      token.length.times do |i|
	assert_equals(token[i], parsed_token[i])
      end
    end
    assert_equals(false, t.next_token.token_type) #there should be no more tokens
  end
  
  def assert_position(position, expected = nil)
    assert_kind_of(Integer, position)
    assert_equals(expected, position) if expected
  end
  
  def assert_tok(string, tokenType, lexeme, restArgs = {})
    prests = restArgs[:prest]
    if not prests.kind_of?(Array)
      prests = [prests ? prests : EXPR_BEG]
    end
    prests.each do |pre_state|
      t = Lexer.new string
      yield t if block_given?
      t.state = pre_state
      token = t.next_token
      assert_kind_of(RaccToken, token)
      assert_equals(tokenType, token[0])
      assert_equals(lexeme, token.lexeme)
      assert_position(token[2], restArgs[:pos])
      assert_equals(t.position, restArgs[:postpos]) if restArgs[:postpos]
      assert_equals(restArgs[:postst], t.state) if restArgs[:postst]
      assert_equals(restArgs[:space_seen], 
		    t.space_seen) if restArgs[:space_seen]
      assert_equals(restArgs[:info], token.info) if restArgs[:info]
      assert_equals(restArgs[:cond_stack], token.cond_stack) if restArgs[:cond_stack] # RACK - needed for parenthesis
      assert_equals(restArgs[:cond_nest], token.cond_nest) if restArgs[:cond_nest] # RACK (see above)
    end
  end
end
